"""
To produce images for different raster grid setups, configure different yaml files
and place them in the config directory as defined by prnu_config.  One environment variable
is necessary for this script to run: ALTIUS_DATABASE_FOLDER, which specifies the location
of the homebase for ALTIUS data and HMI data.

The script produces a PRNU image as computed from the solar-centering algorithm as the default,
a "corrected" PRNU image that minimizes the entropy in the raster scan area of each solar image,
and the "ideal" PRNU image that uses the known center and radius of the solar disk.

The script currently doesn't have a robust method for stitching images together, as it's working
from the assumption that the sun is exactly where it needs to be.
"""
import os
import pickle

import click
import numpy as np

from altius_l1pp_inflight.configuration import altius_config
from altius_l1pp_inflight.prnu.solar.entropy_minimize import find_best_center
from altius_l1pp_inflight.scene.altius_scene import SpsToUsaskSolarSceneInterface, SolarPrnuMapOutput
from altius_l1pp_inflight.scene.disk_geometry import compute_raster_box, center_solar_disk, SolarGeometry
from altius_l1pp_inflight.scene.generation.solar.calibration_util import compare_disk_centers, SolarCalibrationUtil, \
    get_solar_prnu_simulated_observation_dir, get_solar_prnu_calibration_dir
from altius_l1pp_inflight.scene.generation.solar.hmi_files import compute_solar_angular_radius_deg, HmiSunspotScenario
from altius_l1pp_inflight.scene.generation.solar.virtual_solar_detector import RasterScanGrid
from altius_l1pp_inflight.srf.GaussianSrf import GaussianSrf

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.group(context_settings=CONTEXT_SETTINGS)
def cli():
    pass


@cli.command()
@click.option('-i', '--input_prnu_measurement_file', type=str,
              help='Name of .nc file containing the solar measurements for PRNU, '
                   'contained in ALTIUS_L1PP_DATABASE director for solar measurements')
@click.option('-f', '--config_file', type=str,
              help='Name of configuration yaml file stored in ALTIUS_L1PP_DATABASE/config', default=None)
@click.option('-o', '--output_file', type=str,
              help='Output filename to be stored in the ALTIUS_L1PP_DATABASE directory for processed solar PRNU files')
def runsolarprnucal(input_prnu_measurement_file: str, output_file: str, config_filename: str):
    config = altius_config.AltiusL1ppInflightConfig(config_filename, False)
    run_solar_prnu_cal(input_prnu_measurement_file, output_file, config)


def run_solar_prnu_cal(input_prnu_measurement_file: str, output_filename: str,
                       config: altius_config.AltiusL1ppInflightConfig):
    debug = False
    optimize_radius = False
    calibration_wavelength = config.get_solar_prnu_calibration_wavelength()
    home_dir = config.get_processed_solar_prnu_file_directory()
    mode = config.get_detector_mode()
    scenario = HmiSunspotScenario(config.get_solar_scenario())
    channel = config.get_channel(calibration_wavelength)
    pixel_size = config.get_detector_pixel_dimensions(channel=channel, mode=mode)
    horizontal_width = pixel_size[1]
    vertical_width = pixel_size[0]

    num_vert_scans = config.get_num_vertical_raster_scans()
    num_horz_scans = config.get_num_horizontal_raster_scans()
    prnu_cal_image = np.zeros((vertical_width, horizontal_width))
    fit_cal_image = np.zeros_like(prnu_cal_image)

    prnu_cal_image_ideal = np.zeros_like(prnu_cal_image)
    prnu_cal_image_corrected = np.zeros_like(prnu_cal_image)
    prnu_error_est = np.zeros_like(prnu_cal_image)
    prnu_error_est_ideal = np.zeros_like(prnu_cal_image)
    prnu_error_est_corrected = np.zeros_like(prnu_cal_image)

    maxDN = config.get_max_DN()
    max_radiance = config.get_max_radiance()
    factor = maxDN / max_radiance

    srf_wavelen = np.linspace(calibration_wavelength - 25, calibration_wavelength + 25, 5001)  # 0.01 nm res

    srf = GaussianSrf(srf_wavelen, np.ones_like(srf_wavelen) * 3.0)

    fov = config.get_detector_fov(channel=channel)  # (vertical, horizontal)

    satellite_height = config.get_nominal_satellite_height_km()
    vb, hb = config.get_pixel_boresights(channel, mode)

    # Configure the entire solar scan and its geometry:

    output_directory = get_solar_prnu_simulated_observation_dir(config, scenario.to_string_directory())
    if not os.path.exists(output_directory):
        os.makedirs(output_directory)

    num_scans = num_vert_scans * num_horz_scans

    # mean and standard deviation for the image
    # existing_aggregate_hmi = (0, 0, 0)
    # existing_aggregate_altius = (0, 0, 0)

    measurement_times = []

    centering_error = np.zeros((num_horz_scans * num_vert_scans, 3))
    sps_to_usask_interface = SpsToUsaskSolarSceneInterface()

    sps_output_calibration_file = os.path.join(output_directory, input_prnu_measurement_file)

    sps_to_usask_interface.open_read_dataset(sps_output_calibration_file)

    solar_geometry = SolarGeometry(horizontal_width, vertical_width, vb, hb, fov[0], fov[1])

    # TODO might have to do some error-checking to make sure the right file was input

    solar_cal = SolarCalibrationUtil(solar_geometry, srf, 2.0)

    # Compute the ideal/expected raster bounding box:
    altius_scene = sps_to_usask_interface.load_scene(0)
    angular_radius = compute_solar_angular_radius_deg(altius_scene.get_satellite_to_sun_distance_km())
    raster_scan_grid = RasterScanGrid(solar_radius_angular_size_deg=angular_radius,
                                      scan_rate_deg_per_sec=0.1,
                                      time_per_image_sec=1.0,
                                      fov_horizontal_deg=fov[1],
                                      fov_vertical_deg=fov[0],
                                      v_in=num_vert_scans,
                                      h_in=num_horz_scans)

    for raster_idx in range(0, num_scans):

        #############################
        # Data acquisition stage:
        altius_scene = sps_to_usask_interface.load_scene(raster_idx)
        # End data acquisition stage
        #############################

        #############################
        # Processing stage:

        # solar centering:
        image = altius_scene.get_altius_image() / factor
        kwargs = {'detector_angular_width_degrees': altius_scene.get_detector_width_deg(),
                  'distance_to_sun_km': altius_scene.get_satellite_to_sun_distance_km()}
        found_radius, found_horz_center, found_vert_center, ag = center_solar_disk(image, **kwargs)

        centering_error[raster_idx, 0] = altius_scene.get_expected_altius_solar_center_horz() - found_horz_center
        centering_error[raster_idx, 1] = altius_scene.get_expected_altius_solar_center_vert() - found_vert_center
        centering_error[raster_idx, 2] = altius_scene.get_expected_altius_solar_radius() - found_radius

        ###########################
        # Solar correction: TODO this was made more internal?  Need to check...
        # scale = (149.6e6 / altius_scene.get_satellite_to_sun_distance_km())**2
        # print('Image scaled by {0:.3f} to correct for solar distance'.format(scale))
        # image *= scale
        ###########################

        # TODO will need a more robust algorithm for placing pixels on the detector
        #  when things aren't lined up perfectly
        print('Adding raster scan ', raster_idx, ' to PRNU calibration space')

        col, row = raster_scan_grid.get_row_col_meshgrid()

        raster_bounding_box = compute_raster_box(horizontal_width,
                                                 vertical_width,
                                                 raster_scan_grid.grid_horizontal_step_in_pixels(
                                                     altius_scene.get_expected_altius_solar_radius()),
                                                 raster_scan_grid.grid_vertical_step_in_pixels(
                                                     altius_scene.get_expected_altius_solar_radius()),
                                                 raster_scan_grid._padding,
                                                 col[raster_idx],
                                                 row[raster_idx])

        flatten_res = solar_cal.flatten_solar_disk_image(image, found_horz_center, found_vert_center, found_radius,
                                                         altius_scene.get_satellite_to_sun_distance_km(),
                                                         calibration_wavelength, raster_bounding_box)

        flattened_image = flatten_res['flattened_image']
        error_est = flatten_res['error_estimate']

        flatten_res = solar_cal.flatten_solar_disk_image(image, altius_scene.get_expected_altius_solar_center_horz(),
                                                         altius_scene.get_expected_altius_solar_center_vert(),
                                                         found_radius,
                                                         altius_scene.get_satellite_to_sun_distance_km(),
                                                         calibration_wavelength, raster_bounding_box)
        flattened_image_ideal = flatten_res['flattened_image']
        error_est_ideal = flatten_res['error_estimate']

        flat_radiance = solar_cal.flat_solar_disk_model(found_vert_center, found_horz_center,
                                                        found_radius, calibration_wavelength,
                                                        altius_scene.get_satellite_to_sun_distance_km())

        flatten_res = solar_cal.flatten_solar_disk_image(image, altius_scene.get_expected_altius_solar_center_horz(),
                                                         altius_scene.get_expected_altius_solar_center_vert(),
                                                         found_radius,
                                                         altius_scene.get_satellite_to_sun_distance_km(),
                                                         calibration_wavelength, raster_bounding_box)

        flattened_image_ideal = flatten_res['flattened_image']
        error_est_ideal = flatten_res['error_estimate']

        flat_radiance_ideal = solar_cal.flat_solar_disk_model(
            altius_scene.get_expected_altius_solar_center_vert(),
            altius_scene.get_expected_altius_solar_center_horz(),
            altius_scene.get_expected_altius_solar_radius(),
            calibration_wavelength,
            altius_scene.get_satellite_to_sun_distance_km())

        center_result = find_best_center(solar_image=image,
                                         solar_disk_pixel_parameters=(found_vert_center,
                                                                      found_horz_center,
                                                                      found_radius),
                                         raster_pixel_bounds=raster_bounding_box,
                                         wavelength=calibration_wavelength,
                                         solar_geometry=solar_geometry,
                                         distance_to_sun_km=altius_scene.get_satellite_to_sun_distance_km(),
                                         optimize_radius=optimize_radius)

        vert_corrected_center = center_result['vertical_center']
        horz_corrected_center = center_result['horizontal_center']
        corrected_radius = center_result['radius']

        flatten_res = solar_cal.flatten_solar_disk_image(image,
                                                         horz_corrected_center,
                                                         vert_corrected_center,
                                                         corrected_radius,
                                                         altius_scene.get_satellite_to_sun_distance_km(),
                                                         calibration_wavelength,
                                                         raster_bounding_box)

        flattened_image_corrected = flatten_res['flattened_image']
        error_est_corrected = flatten_res['error_estimate']
        flat_radiance_corrected = solar_cal.flat_solar_disk_model(vert_corrected_center, horz_corrected_center,
                                                                  altius_scene.get_expected_altius_solar_radius(),
                                                                  calibration_wavelength,
                                                                  altius_scene.get_satellite_to_sun_distance_km())

        if debug:
            compare_disk_centers(altius_scene, found_horz_center, found_vert_center, found_radius,
                                 horz_corrected_center, vert_corrected_center)

        v_0 = raster_bounding_box[0]
        h_0 = raster_bounding_box[1]
        v_w = raster_bounding_box[2]
        h_w = raster_bounding_box[3]

        prnu_cal_image[v_0:(v_0 + v_w), h_0:(h_0 + h_w)] = flattened_image / flat_radiance
        prnu_cal_image_ideal[v_0:(v_0 + v_w), h_0:(h_0 + h_w)] = flattened_image_ideal / flat_radiance_ideal
        prnu_cal_image_corrected[v_0:(v_0 + v_w), h_0:(h_0 + h_w)] = flattened_image_corrected / flat_radiance_corrected
        prnu_error_est[v_0:(v_0 + v_w), h_0:(h_0 + h_w)] = error_est / flat_radiance
        prnu_error_est_ideal[v_0:(v_0 + v_w), h_0:(h_0 + h_w)] = error_est_ideal / flat_radiance_ideal
        prnu_error_est_corrected[v_0:(v_0 + v_w), h_0:(h_0 + h_w)] = error_est_corrected / flat_radiance_corrected

        # End processing stage
        #############################

    sps_to_usask_interface.close_read_nc()

    error_prnu_image = (prnu_cal_image - prnu_cal_image_ideal) / prnu_cal_image_ideal

    # mean_altius, variance_altius, sample_variance_altius = finalize(existing_aggregate_altius)

    # Write to file:
    prnu_output_directory = get_solar_prnu_calibration_dir(config, scenario.to_string_directory())
    if not os.path.exists(prnu_output_directory):
        os.makedirs(prnu_output_directory)
    prnu_map_out = os.path.join(prnu_output_directory, output_filename)

    SolarPrnuMapOutput(prnu_map_out, calibration_wavelength,
                       prnu_cal_image_corrected, prnu_error_est_corrected)
    print('*********************************************')
    print('Wrote solar PRNU calibration output to ', prnu_map_out, ' for wavelength ', calibration_wavelength, ' nm')
    print('*********************************************')

    # pickle.dump((prnu_cal_image, prnu_error_est),
    #             open(os.path.join(output_directory, 'calibration_image_flat.pickle'), 'wb'))
    #
    # pickle.dump(fit_cal_image,
    #             open(os.path.join(output_directory, 'first_order_fit_calibration_image.pickle'), 'wb'))
    #
    # pickle.dump((prnu_cal_image_ideal, prnu_error_est_ideal),
    #             open(os.path.join(output_directory, 'calibration_image_flat_ideal.pickle'), 'wb'))
    #
    # pickle.dump((prnu_cal_image_corrected, prnu_error_est_corrected),
    #             open(os.path.join(output_directory, 'calibration_image_flat_corrected_center.pickle'), 'wb'))
    #
    # pickle.dump(error_prnu_image,
    #             open(os.path.join(output_directory, 'prnu_error_relative_to_ideal_flattened.pickle'), 'wb'))
    #
    # pickle.dump(centering_error,
    #             open(os.path.join(output_directory, 'centering_error.pickle'), 'wb'))


# Temporary place for CLI
if __name__ == '__main__':
    run_solar_prnu_cal(input_prnu_measurement_file='prnu_solar_measurements.nc',
                       output_filename='prnu_map.nc', config=altius_config.AltiusL1ppInflightConfig())
