import os

import click
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr

from altius_l1pp_inflight.configuration import altius_config
from altius_l1pp_inflight.prnu.solar.produce_calibration_image import run_solar_prnu_cal
from altius_l1pp_inflight.scene.altius_scene import SpsToUsaskSolarSceneInterface, SolarPrnuMapOutput
from altius_l1pp_inflight.scene.disk_geometry import SolarGeometry, D_1AU, SOLAR_RADIUS
from altius_l1pp_inflight.scene.generation.solar.calibration_util import SolarCalibrationUtil, \
    get_solar_prnu_simulated_observation_dir, get_solar_prnu_calibration_dir
from altius_l1pp_inflight.scene.generation.solar.hmi_files import HmiSunspotScenario
from altius_l1pp_inflight.scene.generation.solar.virtual_solar_detector import RasterScanGrid
from altius_l1pp_inflight.srf.GaussianSrf import GaussianSrf

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.group(context_settings=CONTEXT_SETTINGS)
def cli():
    pass


@cli.command()
def runsolarprnutest():
    run_solar_prnu_test()


def run_solar_prnu_test(show_figure=False):
    config = altius_config.AltiusL1ppInflightConfig(test_config=True)

    num_vert_scans = config.get_num_vertical_raster_scans()
    num_horz_scans = config.get_num_horizontal_raster_scans()

    maxDN = config.get_max_DN()
    max_radiance = config.get_max_radiance()
    factor = maxDN / max_radiance

    pixel_size = config.get_detector_pixel_dimensions(channel='vis', mode='native')
    horizontal_width = pixel_size[1]
    vertical_width = pixel_size[0]
    wavelength = 500

    fov = config.get_detector_fov('vis')  # (vertical, horizontal)
    vb, hb = config.get_pixel_boresights('vis', 'native')

    srf_wavelen = np.linspace(wavelength - 25, wavelength + 25, 5001)  # 0.01 nm res

    srf = GaussianSrf(srf_wavelen, np.ones_like(srf_wavelen) * 3.0)

    solar_geometry = SolarGeometry(horizontal_width, vertical_width,
                                   vb, hb,
                                   fov[0], fov[1])

    solar_cal = SolarCalibrationUtil(solar_geometry, srf, 1.2)

    sps_to_usask_interface = SpsToUsaskSolarSceneInterface()

    x = 2 * D_1AU * np.tan(np.deg2rad(fov[1]) / 2)
    r = SOLAR_RADIUS * horizontal_width / x

    output_directory = get_solar_prnu_simulated_observation_dir(config,
                                                                HmiSunspotScenario.NO_SUNSPOTS.to_string_directory())
    test_file = os.path.join(output_directory, 'test_ideal_solar_images.nc')

    sps_to_usask_interface.open_file_for_write(test_file, wavelength, fov, pixel_size, (vb, hb))

    raster_scan_grid = RasterScanGrid(solar_radius_angular_size_deg=0.5,
                                      scan_rate_deg_per_sec=0.1,
                                      time_per_image_sec=1.0,
                                      fov_horizontal_deg=fov[1],
                                      fov_vertical_deg=fov[0],
                                      v_in=num_vert_scans,
                                      h_in=num_horz_scans)

    cols, rows = raster_scan_grid.get_row_col_meshgrid()

    for row, col in zip(rows, cols):
        cv = (row + 0.5) * vertical_width / num_vert_scans

        ch = (col + 0.5) * horizontal_width / num_horz_scans
        image = solar_cal.model_noiseless_solar_disk(cv, ch, r, wavelength, D_1AU, False)

        image *= factor
        sps_to_usask_interface.write_next_solar_scene_to_netcdf(np.datetime64('1999-12-31'),
                                                                image,
                                                                wavelength,
                                                                (cv, ch),
                                                                r,
                                                                D_1AU,
                                                                np.array([1, 0, 0]),
                                                                np.array([1, 0, 0]),
                                                                0.1)
        print('Completed solar image [{}, {}]'.format(row, col))

    run_solar_prnu_cal(input_prnu_measurement_file='test_ideal_solar_images.nc',
                       output_filename='prnu_test_output.nc', config=config)

    prnu_dir = get_solar_prnu_calibration_dir(config, HmiSunspotScenario.NO_SUNSPOTS.to_string_directory())
    prnu_file = os.path.join(prnu_dir, 'prnu_test_output.nc')

    ds = xr.open_dataset(prnu_file)

    vals = ds.data_vars.variables
    prnu_map = vals[SolarPrnuMapOutput.PRNU_MAP].values

    ones = np.ones_like(prnu_map)

    if show_figure:
        plt.imshow(prnu_map)
        plt.colorbar()
        plt.show()

    assert np.isclose(prnu_map, ones, rtol=0.003).all(), 'Some pixels are more than 0.3% from expected'
    print('*****************************************')
    print('Ideal solar image PRNU passed within 0.3%')
    print('*****************************************')


if __name__ == '__main__':
    run_solar_prnu_test()
