from typing import Tuple

import netCDF4 as nc
import numpy as np
import sktime.mjd as mjd_conv
import xarray as xr

from altius_l1pp_inflight.configuration import altius_config
from altius_l1pp_inflight.scene.disk_geometry import SolarGeometry

"""
This module will include different ways to describe the ALTIUS
scenes used for calibration.  At their core they contain a 2D
array of DN values, and include the effects of the SRF


Set of classes and methods for describing an image taken by the
ALTIUS detector.  Still a work in progress.


TODO Can be used to build further classes that show different stages of calibration correction
"""


class AltiusSolarScene:

    def __init__(self,
                 altius_image,
                 detector_width_deg,
                 satellite_to_sun_distance_km,
                 expected_altius_solar_center_horz,
                 expected_altius_solar_center_vert,
                 expected_altius_solar_radius,
                 wavelength,
                 boresight,
                 integration_time,
                 start_pointing_vector,
                 end_pointing_vector):
        self._altius_image = altius_image
        self._detector_width_deg = detector_width_deg
        self._satellite_to_sun_distance_km = satellite_to_sun_distance_km
        self._expected_altius_solar_center_horz = expected_altius_solar_center_horz
        self._expected_altius_solar_center_vert = expected_altius_solar_center_vert
        self._expected_altius_solar_radius = expected_altius_solar_radius
        self._wavelength = wavelength
        self._boresight = boresight
        self._integration_time = integration_time
        self._starting_pointing_vector = start_pointing_vector
        self._end_pointing_vector = end_pointing_vector

    def get_start_boresight_pointing_vector(self):
        return self._starting_pointing_vector

    def get_end_boresight_pointing_vector(self):
        return self._end_pointing_vector

    def get_integration_time(self):
        return self._integration_time

    def get_boresight(self):
        return self._boresight

    def get_wavelength(self):
        return self._wavelength

    def get_altius_image(self):
        return self._altius_image

    def get_detector_width_deg(self):
        return self._detector_width_deg

    def get_satellite_to_sun_distance_km(self):
        return self._satellite_to_sun_distance_km

    def get_expected_altius_solar_center_horz(self):
        return self._expected_altius_solar_center_horz

    def get_expected_altius_solar_center_vert(self):
        return self._expected_altius_solar_center_vert

    def get_expected_altius_solar_radius(self):
        return self._expected_altius_solar_radius

    def compute_irradiance(self, horizontal_center, vertical_center, conversion_factor, mask_radius=None):
        """
            Compute the irradiance of this solar disk for the assumed center of the sun,
            and with an optional mask applied to only get a portion of the disk.
        """

        h = np.arange(0.5, len(self._altius_image), step=1.0)
        v = np.arange(0.5, len(self._altius_image[0]), step=1.0)
        solar_geometry = SolarGeometry(len(self._altius_image[0]), len(self._altius_image), self._boresight[0],
                                       self._boresight[1])
        solar_geometry.set_disk(self._expected_altius_solar_radius, self._expected_altius_solar_center_horz,
                                self._expected_altius_solar_center_vert, self._satellite_to_sun_distance_km)
        delta_omega = solar_geometry.delta_omega_view()
        rad = np.copy(self._altius_image)
        vv, hh = np.meshgrid(h, v)
        pixel_radii = np.sqrt((vv - vertical_center) ** 2 + (hh - horizontal_center) ** 2)

        if mask_radius is not None:
            mask = pixel_radii <= mask_radius
        else:
            mask = pixel_radii <= self._expected_altius_solar_radius

        rad *= mask

        return np.nansum(rad * delta_omega) * conversion_factor, mask


class AltiusLunarScene:
    def __init__(self,
                 altius_image,
                 detector_width_deg,
                 satellite_to_moon_distance_km,
                 expected_altius_lunar_center_horz,
                 expected_altius_lunar_center_vert,
                 expected_altius_lunar_radius,
                 wavelength,
                 boresight,
                 integration_time,
                 start_pointing_vector,
                 end_pointing_vector,
                 sun_moon_distance_au,
                 lunar_phase_angle_deg,
                 subsatellite_selenographic_latitude,
                 subsatellite_selenographic_longitude,
                 subsolar_selenographic_longitude):
        self._subsolar_selenographic_longitude = subsolar_selenographic_longitude
        self._subsatellite_selenographic_longitude = subsatellite_selenographic_longitude
        self._subsatellite_selenographic_latitude = subsatellite_selenographic_latitude
        self._lunar_phase_angle_deg = lunar_phase_angle_deg
        self._moon_sun_distance_au = sun_moon_distance_au
        self._altius_image = altius_image
        self._detector_width_deg = detector_width_deg
        self._satellite_to_moon_distance_km = satellite_to_moon_distance_km
        self._expected_altius_lunar_center_horz = expected_altius_lunar_center_horz
        self._expected_altius_lunar_center_vert = expected_altius_lunar_center_vert
        self._expected_altius_lunar_radius = expected_altius_lunar_radius
        self._wavelength = wavelength
        self._boresight = boresight
        self._integration_time = integration_time
        self._starting_pointing_vector = start_pointing_vector
        self._end_pointing_vector = end_pointing_vector

    def get_start_boresight_pointing_vector(self):
        return self._starting_pointing_vector

    def get_end_boresight_pointing_vector(self):
        return self._end_pointing_vector

    def get_integration_time(self):
        return self._integration_time

    def get_boresight(self):
        return self._boresight

    def get_wavelength(self):
        return self._wavelength

    def get_altius_image(self):
        return self._altius_image

    def get_detector_width_deg(self):
        return self._detector_width_deg

    def get_satellite_to_moon_distance_km(self):
        return self._satellite_to_moon_distance_km

    def get_expected_altius_lunar_center_horz(self):
        return self._expected_altius_lunar_center_horz

    def get_expected_altius_lunar_center_vert(self):
        return self._expected_altius_lunar_center_vert

    # TODO some work needs to be done with the lunar center, since it's not perfectly circular
    def get_expected_altius_lunar_radius(self):
        return self._expected_altius_lunar_radius

    def compute_irradiance(self, horizontal_center, vertical_center, conversion_factor):
        """
            Compute the irradiance of this lunar disk for the assumed center of the moon,
            and with an optional mask applied to only get a portion of the disk.
        """

        h = np.arange(0.5, len(self._altius_image), step=1.0)
        v = np.arange(0.5, len(self._altius_image[0]), step=1.0)
        solar_geometry = SolarGeometry(len(self._altius_image[0]), len(self._altius_image), self._boresight[0],
                                       self._boresight[1])
        solar_geometry.set_disk(self._expected_altius_lunar_radius, self._expected_altius_lunar_center_horz,
                                self._expected_altius_lunar_center_vert, self._satellite_to_moon_distance_km)
        delta_omega = solar_geometry.delta_omega_view()

        rad = np.copy(self._altius_image)
        vv, hh = np.meshgrid(h, v)
        pixel_radii = np.sqrt((vv - vertical_center) ** 2 + (hh - horizontal_center) ** 2)

        mask = pixel_radii <= self._expected_altius_lunar_radius

        rad *= mask

        return np.nansum(rad * delta_omega) * conversion_factor, mask

    def get_moon_sun_distance_au(self):
        return self._moon_sun_distance_au

    def get_lunar_phase_angle_deg(self):
        return self._lunar_phase_angle_deg

    def get_subsatellite_selenographic_latitude(self):
        return self._subsatellite_selenographic_latitude

    def get_subsatellite_selenographic_longitude(self):
        return self._subsatellite_selenographic_longitude

    def get_subsolar_selenographic_longitude(self):
        return self._subsolar_selenographic_longitude


class SpsHighResolutionUsaskToSpsSceneInterface:
    """
    Class for passing high-resolution (4096x4096) simulated images of the sun to the
    SPS.  TBD if QinetiQ wants this or just the detector resolution images -
    if they want to simulate satellite pointing drift, it's likely they'll want this instead
    of a pre-computed SpsSceneInterface at detector resolution

    Images of the sun are produced from the solar HMI database.  Originally, images were directly
    taken from the HMI database because they accurately show solar convection cells and how they evolve
    with time.  However, it was determined the HMI data contained a certain bias, so solar images were
    decided to be simulated.  Realistic simulation of the solar convection cells is achieved by
    adding Gaussian noise of the same SNR as the HMI images to a limb-darkened solar disk.

    The set of images in this file usually consist of ~100 solar images, which is about 4 GB.

    Dimensions:

    - TIME_DIM
        - value: 'time'
        - variable length dimension - depends on number of images passed in
    - NUM_VERT_PIXELS
        - value: 'num_vert_pixels'
        - fixed number of vertical pixels in the detector
    - NUM_HORZ_PIXELS
        - value: 'num_horz_pixels'
        - fixed number of horizontal pixels in the detector
    - PIXEL_LOCATION_DIM
        - value: 'pixel_location_dim'
        - vert, horz coordinates of any pixel
    - ICF_VECTOR_DIM
        - value: 'icf_vector_dim'
        - xyz of ICF vector
    - NUM_SPECTRUM_WAVELENGTHS
        - value: 'num_spectrum_wavelengths'
        - length of solar spectrum used

    Variables:

    - WAVELENGTHS
        - value: 'wavelengths'
        - size: [TIME_DIM,]
        - wavelength in nm of each simulation
    - HI_RES_SOLAR_DISKS
        - value: 'scenes'
        - size: [TIME_DIM, NUM_VERT_PIXELS, NUM_HORZ_PIXELS]
        - Simulated image of the sun at high spatial resolution, for each interpolated time.  In arbitrary units
    - TIMES
        - value: 'times'
        - size: [TIME_DIM,]
        - Time of each image measurement in MJD
    - IMAGE_SOLAR_CENTER
        - value: 'expected_solar_center'
        - size: [TIME_DIM, PIXEL_LOCATION_DIM]
        - Vertical, horizontal pixel location of where the solar disk is centered.  The simulations
          generally place the center in the middle of the array, [2047.5, 2047.5]
    - IMAGE_SOLAR_RADIUS
        - value: 'expected_solar_radius'
        - size: [TIME_DIM,]
        - Radius of the solar disk for each image.
    - DETECTOR_FOV_DEG
        - value: 'detector_fov_deg'
        - size: [PIXEL_LOCATION_DIM,]
        - Vertical and horizontal field of view in degrees on the detector
    - SOLAR_DISK_CENTER_ICF
        - value: 'solar_disk_center_icf'
        - size: [TIME_DIM, ICF_VECTOR_DIM]
        - Location of the center of the solar disk on the detector in ICF coordinates
    - SOLAR_SPECTRUM_WAVELENGTHS
        - value: 'solar_spectrum_wavelengths'
        - size: [NUM_SPECTRUM_WAVELENGTHS,]
        - wavelengths, in nm of the solar spectrum to be used to convolve with the SRF when scaling the solar disk
    - SOLAR_SPECTRUM
        - value: 'solar_spectrum'
        - size: [NUM_SPECTRUM_WAVELENGTHS,]
        - solar spectrum values in photons/s/cm2/nm
    """
    # Dimensions:
    TIME_DIM = 'time'
    NUM_VERT_PIXELS = 'num_vert_pixels'
    NUM_HORZ_PIXELS = 'num_horiz_pixels'
    PIXEL_LOCATION_DIM = 'pixel_location_dim'
    ICF_VECTOR_DIM = 'icf_vector_dim'
    NUM_SPECTRUM_WAVELENGTHS = 'num_spectrum_wavelengths'

    # Variable names:
    WAVELENGTHS = 'wavelengths'
    HI_RES_SOLAR_DISKS = 'scenes'
    TIMES = 'times'
    IMAGE_SOLAR_CENTER = 'image_solar_center'
    IMAGE_SOLAR_RADIUS = 'image_solar_radius'
    DETECTOR_FOV_DEG = 'detector_fov_deg'
    SOLAR_DISK_CENTER_ICF = 'solar_disk_center_icf'
    SOLAR_SPECTRUM_WAVELENGTHS = 'solar_spectrum_wavelengths'
    SOLAR_SPECTRUM = 'solar_spectrum'

    def __init__(self):
        self._write_nc_file = None

    def open_file_for_write(self,
                            filename: str,
                            detector_fov: Tuple[float, float],
                            solar_spectrum_wavelengths: np.ndarray,
                            solar_spectrum: np.ndarray):
        """
        Create a new netcdf file to write images to.

        Parameters
        ----------
        filename: str
            Full path to the netcdf file to write.

        detector_fov: Tuple[float, float]
            Vertical, horizontal FOV of the detector in degrees.

        solar_spectrum_wavelengths: np.ndarray
            Wavelengths, in nm, where the solar spectrum is evaluated.

        solar_spectrum: np.ndarray
            Irradiance spectrum of the sun at solar_spectrum_wavelengths, in photons/s/cm2/nm
        """
        self._write_nc_file = nc.Dataset(filename, mode='w', format='NETCDF4')

        # Dimensions:
        self._time_dimension = self._write_nc_file.createDimension(self.TIME_DIM, None)  # variable size (appending)
        self._pixel_location_dim = self._write_nc_file.createDimension(self.PIXEL_LOCATION_DIM, 2)
        self._icf_dim = self._write_nc_file.createDimension(self.ICF_VECTOR_DIM, 3)
        self._num_vert_pixels = self._write_nc_file.createDimension(self.NUM_VERT_PIXELS, 4096)
        self._num_horz_pixels = self._write_nc_file.createDimension(self.NUM_HORZ_PIXELS, 4096)

        self._wavelengths = self._write_nc_file.createVariable(self.WAVELENGTHS, float, (self.TIME_DIM,))
        self._wavelengths.units = 'nm'

        self._detector_fov = self._write_nc_file.createVariable(self.DETECTOR_FOV_DEG, float,
                                                                (self.PIXEL_LOCATION_DIM,))
        self._detector_fov.units = 'deg'
        self._detector_fov.description = 'Vertical and horizontal angular width of detector, respectively'
        self._detector_fov[:] = [detector_fov[0], detector_fov[1]]

        self._num_spectrum_wavelengths = self._write_nc_file.createDimension(self.NUM_SPECTRUM_WAVELENGTHS,
                                                                             len(solar_spectrum_wavelengths))

        self._solar_spectrum_wavelengths = self._write_nc_file.createVariable(self.SOLAR_SPECTRUM_WAVELENGTHS,
                                                                              float,
                                                                              (self.NUM_SPECTRUM_WAVELENGTHS,))
        self._solar_spectrum_wavelengths.units = 'nm'
        self._solar_spectrum_wavelengths.description = 'Wavelengths (in nm) of the solar spectrum to be ' \
                                                       'used in scaling the solar disk'
        self._solar_spectrum_wavelengths[:] = solar_spectrum_wavelengths

        self._solar_spectrum = self._write_nc_file.createVariable(self.SOLAR_SPECTRUM,
                                                                  float,
                                                                  (self.NUM_SPECTRUM_WAVELENGTHS,))
        self._solar_spectrum.units = 'photons/s/cm2/nm'
        self._solar_spectrum.description = 'Total disk irradiance spectrum of the sun at 1 AU used for ' \
                                           'scaling the solar disk'
        self._solar_spectrum[:] = solar_spectrum

        self._solar_disks = self._write_nc_file.createVariable(self.HI_RES_SOLAR_DISKS, float,
                                                               (self.TIME_DIM,
                                                                self.NUM_VERT_PIXELS,
                                                                self.NUM_HORZ_PIXELS),
                                                               zlib=True, complevel=9,
                                                               least_significant_digit=6)
        self._solar_disks.units = 'arbitrary'
        self._solar_disks.description = 'High-resolution solar image with limb darkening applied at wavelength'

        self._times = self._write_nc_file.createVariable(self.TIMES, float, (self.TIME_DIM,))
        self._times.units = 'MJD (seconds since 1970-01-01)'

        self._image_solar_center = self._write_nc_file.createVariable(self.IMAGE_SOLAR_CENTER, float,
                                                                      (self.TIME_DIM, self.PIXEL_LOCATION_DIM))
        self._image_solar_center.description = 'Center of the solar disk in vertical ' \
                                               'and horizontal pixels, respectively'
        self._image_solar_center.units = 'HMI Pixels'
        self._image_disk_radius = self._write_nc_file.createVariable(self.IMAGE_SOLAR_RADIUS, float,
                                                                     (self.TIME_DIM,))
        self._image_disk_radius.units = 'HMI Pixels'

        self._solar_detector_location_icf = self._write_nc_file.createVariable(self.SOLAR_DISK_CENTER_ICF, float,
                                                                               (self.TIME_DIM, self.ICF_VECTOR_DIM))
        self._solar_detector_location_icf.description = 'Unit vector in ICF coordinates ' \
                                                        'pointing to the center of the solar disk on the detector'
        self._solar_detector_location_icf.units = 'ICF unit vector'

        self._current_index = 0

    def write_next_solar_disk_to_netcdf(self,
                                        time: np.datetime64,
                                        high_res_disk_image: np.ndarray,
                                        wavelength: float,
                                        disk_center: Tuple[float, float],
                                        disk_radius_pixels: float,
                                        icf_solar_center: np.ndarray):
        """
        Parameters
        ----------
        time: np.datetime64
            Time of the measurement over the course of the orbit.

        high_res_disk_image: np.ndarray
            Image of the sun (4096x4096), in arbitrary units

        wavelength: float
            Wavelength of the simulation in nm.

        disk_center: Tuple[float, float]
            Vertical, horizontal pixel center location of the disk.

        disk_radius_pixels: float
            Pixel radius of solar disk.

        icf_solar_center: np.ndarray
            ICF detector location of the solar disk center.
        """
        self._wavelengths[self._current_index] = wavelength
        self._times[self._current_index] = mjd_conv.datetime64_to_mjd(time)
        self._solar_disks[self._current_index, :, :] = high_res_disk_image
        self._image_solar_center[self._current_index, :] = [disk_center[0], disk_center[1]]
        self._image_disk_radius[self._current_index] = disk_radius_pixels
        self._solar_detector_location_icf[self._current_index, :] = icf_solar_center

        self._current_index += 1

    def close(self):
        self._write_nc_file.close()


class SpsToUsaskSolarSceneInterface:
    """
    This class represents the output from the SPS given to the calibration
    processes using images of the sun.  So far this includes the PRNU and SRF calibrations.
    It contains images of the sun for a single wavelength calibration.

    For a PRNU calibration, a single instance of the class will contain images of the sun
    at a single wavelength over all locations on the detector.  This usually consists of ~100
    images, typically at detector resolution, resulting in a ~1 GB file.

    For an SRF calibration, a single instance of the class will contain multiples wavelengths
    of the sun at a single location.  The number of wavelengths will depend on future studies to
    determine the number of wavelengths necessary for an accurate calibration.  If a small number are
    needed, then multiple sun locations will be included, but this is TBD.  Either way, the intent
    is that the measurements are intended for calibrating a single wavelength at the center
    of the measurement wavelengths.

    The contents of the class are saved in netcdf format.

    Dimensions:

    - TIME_DIM
        - value: 'time'
        - variable length dimension - depends on number of images passed in
    - NUM_VERT_PIXELS
        - value: 'num_vert_pixels'
        - fixed number of vertical pixels in the detector
    - NUM_HORZ_PIXELS
        - value: 'num_horz_pixels'
        - fixed number of horizontal pixels in the detector
    - PIXEL_LOCATION_SIZE
        - value: 'pixel_location_size'
        - vert, horz coordinates of any pixel
    - CAL_WL_DIM
        - value: 'calibration_wavelength_nm'
        - length of the calibration wavelength (1)
    - BORESIGHT_VECTOR_DIM
        - value: 'boresight_vector_dim'
        - length of the vector for specifying look direction of the boresight (3)

    Variables:

    - DETECTOR_BORESIGHT
        - value: 'detector_boresight'
        - size: [PIXEL_LOCATION_SIZE,]
        - vertical, horizontal pixel location of the detector boresight
    - WAVELENGTHS
        - value: 'wavelengths'
        - size: [TIME_DIM,]
        - wavelength in nm of each measurement
    - SCENES
        - value: 'scenes'
        - size: [TIME_DIM, NUM_VERT_PIXELS, NUM_HORZ_PIXELS]
        - Image of the sun on the detector for each measurement.  In units of DN
    - TIMES
        - value: 'times'
        - size: [TIME_DIM,]
        - Time of each image measurement in MJD
    - EXPECTED_SOLAR_CENTER
        - value: 'expected_solar_center'
        - size: [TIME_DIM, PIXEL_LOCATION_SIZE]
        - Vertical, horizontal pixel location of where the solar disk was intended to be
          aimed on the detector surface.  The actual position of the sun may differ due
          to pointing errors.
    - EXPECTED_SOLAR_RADIUS
        - value: 'expected_solar_radius'
        - size: [TIME_DIM,]
        - Expected radius of the solar disk on the detector surface in pixels.  Can be directly computed
          based on the solar distance and constant solar radius.  This is only used to have a starting
          point for finding the actual solar radius of the image.
    - SAT_TO_SUN_DISTANCE
        - value: 'satellite_to_sun_distance'
        - size: [TIME_DIM, ]
        - Distance from the satellite to the sun at each measurement, in km.
    - DETECTOR_FOV_DEG
        - value: 'detector_fov_deg'
        - size: [PIXEL_LOCATION_SIZE,]
        - Vertical and horizontal field of view in degrees on the detector
    - CALIBRATION_WAVELENGTH
        - value: 'calibration_wavelength_nm'
        - size: [CAL_WL_DIM,]
        - The calibration wavelength in nm for the measurements
    - START_SNAPSHOT_BORESIGHT_POINTING
        - value: 'start_snapshot_pointing'
        - size: [TIME_DIM, BORESIGHT_VECTOR_DIM]
        - Vector of the boresight in (TBD coordinates) at the start of the integration for each measurement.
          Used to help correct for the blurring that occurs.
    - END_SNAPSHOT_BORESIGHT_POINTING
        - value: 'end_snapshot_pointing'
        - size: [TIME_DIM, BORESIGHT_VECTOR_DIM]
        - Vector of the boresight in (TBD coordinates) at the end of the integration for each measurement.
          Used to help correct for the blurring that occurs.
    - INTEGRATION_TIMES
        - value: 'integration_times'
        - size: [TIME_DIM,]
        - Integration time in seconds of each measurement.
    """

    # Dimensions:
    TIME_DIM = 'time'
    NUM_VERT_PIXELS = 'num_vert_pixels'
    NUM_HORZ_PIXELS = 'num_horz_pixels'
    PIXEL_LOCATION_SIZE = 'pixel_location_size'
    CAL_WL_DIM = 'calibration_wavelength_dimension'
    BORESIGHT_VECTOR_DIM = 'boresight_vector_dim'

    # Variable names:
    DETECTOR_BORESIGHT = 'detector_boresight'
    WAVELENGTHS = 'wavelengths'
    SCENES = 'scenes'
    TIMES = 'times'
    EXPECTED_SOLAR_CENTER = 'expected_solar_center'
    EXPECTED_SOLAR_RADIUS = 'expected_solar_radius'
    SAT_TO_SUN_DISTANCE = 'satellite_to_sun_distance'
    DETECTOR_FOV_DEG = 'detector_fov_deg'
    CALIBRATION_WAVELENGTH = 'calibration_wavelength_nm'
    START_BORESIGHT_POINTING = 'start_snapshot_pointing'
    END_BORESIGHT_POINTING = 'end_snapshot_pointing'
    INTEGRATION_TIMES = 'integration_times'

    def __init__(self):
        self._read_nc_file = None
        self._write_nc_file = None

    def open_file_for_write(self,
                            filename: str,
                            calibration_wavelength: float,
                            detector_fov: Tuple[float, float],
                            detector_size_pixels: Tuple[int, int],
                            detector_boresight_pixels: Tuple[float, float]):
        """
        Initialize a new netcdf file for writing data.

        Parameters
        ----------
        filename : str
            Full path string to the netcdf file.

        calibration_wavelength : float
            Wavelength in nm for the calibration measurements

        detector_fov : Tuple[float, float]
            Vertical, horizontal FOV of the detector, in degrees

        detector_size_pixels : Tuple[int, int]
            Vertical, horizontal number of pixels for the detector

        detector_boresight_pixels : Tuple[float, float]
            Vertical, horizontal pixel location of the boresight of the detector.

        """

        self._write_nc_file = nc.Dataset(filename, mode='w', format='NETCDF4')

        # Dimensions:
        self._time_dimension = self._write_nc_file.createDimension(self.TIME_DIM, None)  # variable size (appending)
        self._num_vert_pixels = self._write_nc_file.createDimension(self.NUM_VERT_PIXELS, detector_size_pixels[0])
        self._num_horz_pixels = self._write_nc_file.createDimension(self.NUM_HORZ_PIXELS, detector_size_pixels[1])
        self._pixel_location_size = self._write_nc_file.createDimension(self.PIXEL_LOCATION_SIZE, 2)
        self._calibration_wl_dim = self._write_nc_file.createDimension(self.CAL_WL_DIM, 1)
        self._boresight_vector_dim = self._write_nc_file.createDimension(self.BORESIGHT_VECTOR_DIM, 3)

        # Variables:
        self._snapshot_pointing_start = self._write_nc_file.createVariable(self.START_BORESIGHT_POINTING,
                                                                           float,
                                                                           (self.TIME_DIM, self.BORESIGHT_VECTOR_DIM))
        self._snapshot_pointing_start.units = 'Arbitrary'
        self._snapshot_pointing_start.description = 'ECEF absolute vector describing the pointing of the ' \
                                                    'instrument boresight at the beginning of each integration period'

        self._snapshot_pointing_end = self._write_nc_file.createVariable(self.END_BORESIGHT_POINTING,
                                                                         float,
                                                                         (self.TIME_DIM, self.BORESIGHT_VECTOR_DIM))
        self._snapshot_pointing_end.units = 'Arbitrary'
        self._snapshot_pointing_end.description = 'ECEF absolute vector describing the pointing of the instrument ' \
                                                  'boresight at the end of each integration period'

        self._integration_times = self._write_nc_file.createVariable(self.INTEGRATION_TIMES, float, (self.TIME_DIM,))
        self._integration_times.units = 'seconds'
        self._integration_times.description = 'Exposure time of each snapshot of the sun'

        self._detector_boresight = self._write_nc_file.createVariable(self.DETECTOR_BORESIGHT, float,
                                                                      (self.PIXEL_LOCATION_SIZE,))
        self._detector_boresight.units = 'Pixels'
        self._detector_boresight[:] = [detector_boresight_pixels[0], detector_boresight_pixels[1]]

        self._calibration_wavelength = self._write_nc_file.createVariable(self.CALIBRATION_WAVELENGTH,
                                                                          float, (self.CAL_WL_DIM,))
        self._calibration_wavelength[:] = calibration_wavelength
        self._calibration_wavelength.units = 'nm'
        self._calibration_wavelength.description = 'Wavelength in nm for which the set of images is intended ' \
                                                   'to calibrate'

        self._wavelengths = self._write_nc_file.createVariable(self.WAVELENGTHS, float, (self.TIME_DIM,))
        self._wavelengths.units = 'nm'

        self._detector_fov = self._write_nc_file.createVariable(self.DETECTOR_FOV_DEG, float,
                                                                (self.PIXEL_LOCATION_SIZE,))
        self._detector_fov.units = 'deg'
        self._detector_fov.description = 'Vertical and horizontal angular width of detector, respectively'
        self._detector_fov[:] = [detector_fov[0], detector_fov[1]]

        self._scenes = self._write_nc_file.createVariable(self.SCENES, float,
                                                          (self.TIME_DIM,
                                                           self.NUM_VERT_PIXELS,
                                                           self.NUM_HORZ_PIXELS))
        self._scenes.units = 'DN'
        self._scenes.description = 'Scene at detector resolution'

        self._times = self._write_nc_file.createVariable(self.TIMES, float, (self.TIME_DIM,))
        self._times.units = 'MJD'

        self._expected_solar_center = self._write_nc_file.createVariable(self.EXPECTED_SOLAR_CENTER, float,
                                                                         (self.TIME_DIM, self.PIXEL_LOCATION_SIZE))
        self._expected_solar_center.description = 'Center of the solar disk in vertical and horizontal pixels, ' \
                                                  'respectively'
        self._expected_solar_center.units = 'Pixels'

        self._expected_solar_radius = self._write_nc_file.createVariable(self.EXPECTED_SOLAR_RADIUS, float,
                                                                         (self.TIME_DIM,))
        self._expected_solar_radius.units = 'Pixels'

        self._satellite_to_sun_distance = self._write_nc_file.createVariable(self.SAT_TO_SUN_DISTANCE, float,
                                                                             (self.TIME_DIM,))
        self._satellite_to_sun_distance.units = 'km'

        self._current_index = 0

    def write_next_solar_scene_to_netcdf(self,
                                         time: np.datetime64,
                                         scene: np.ndarray,
                                         wavelength: float,
                                         expected_solar_center: Tuple[float, float],
                                         expected_solar_radius: float,
                                         sat_sun_distance_km: float,
                                         start_boresight_pointing: np.ndarray,
                                         end_boresight_pointing: np.ndarray,
                                         integration_time: float):
        """
        Once a netcdf file is open for writing, calling this appends another detector
        image to the set of images in currently open netcdf.

        Parameters
        ----------
        time : np.datetime64
            Time at the start of the measurement.

        scene : np.ndarray
            The actual detector pixels with measured DN values of the sun.

        wavelength : float
            Wavelength in nm of the measurement, not necessarily the same as the calibration wavelength.

        expected_solar_center : Tuple[float, float]
            Vertical, horizontal pixel location of the location of the sun where it was intended to be.

        expected_solar_radius : float
            Pixel radius of the solar disk as expected by the detector FOV and solar distance.

        sat_sun_distance_km : float
            Distance from the satellite to the sun in km.

        start_boresight_pointing : np.ndarray
            xyz vector of the boresight in TBD coordinate system at the start of the integration time.

        end_boresight_pointing: np.ndarray
            xyz vector of the boresight in TBD coordinate system at the end of the integration time.

        integration_time : float
            Integration time of the measurement in seconds.
        """
        self._wavelengths[self._current_index] = wavelength
        self._times[self._current_index] = mjd_conv.datetime64_to_mjd(time)
        self._scenes[self._current_index, :, :] = scene
        self._expected_solar_center[self._current_index, :] = [expected_solar_center[0], expected_solar_center[1]]
        self._expected_solar_radius[self._current_index] = expected_solar_radius
        self._satellite_to_sun_distance[self._current_index] = sat_sun_distance_km
        self._snapshot_pointing_start[self._current_index, :] = start_boresight_pointing
        self._snapshot_pointing_end[self._current_index, :] = end_boresight_pointing
        self._integration_times[self._current_index] = integration_time

        self._current_index += 1

    def close_write_nc(self):
        self._write_nc_file.close()

    def close_read_nc(self):
        if self._read_nc_file is not None:
            self._read_nc_file.close()

    def open_read_dataset(self, filename):
        """
        Open this netcdf file for reading.

        Parameters
        ----------
        filename : str
            Path to the netcdf file.
        """
        self._read_nc_file = xr.open_dataset(filename)  # , chunks={SpsToUsaskSceneInterface.TIME_DIM: 4})

    def get_num_scenes(self):
        return self._read_nc_file.dims[SpsToUsaskSolarSceneInterface.TIME_DIM]

    def load_scene(self, scene_index) -> AltiusSolarScene:
        """
        Get an AltiusSolarScene instance for the given scene index from the current file.
        Must have called :code:`open_read_dataset` prior to this, and can use
        :code:`get_num_scenes` to figure out how many scenes are present.

        Parameters
        ----------
        scene_index : int
            Raster scene index of the solar scene wanted from the netcdf file.

        Returns
        -------
        val : AltiusSolarScene
            An instance of the AltiusSolarScene for this measurement.
        """
        vals = self._read_nc_file.variables
        expected_center = vals[SpsToUsaskSolarSceneInterface.EXPECTED_SOLAR_CENTER].values[scene_index]
        return AltiusSolarScene(vals[SpsToUsaskSolarSceneInterface.SCENES].values[scene_index],
                                vals[SpsToUsaskSolarSceneInterface.DETECTOR_FOV_DEG].values,
                                vals[SpsToUsaskSolarSceneInterface.SAT_TO_SUN_DISTANCE].values[scene_index],
                                expected_center[1],
                                expected_center[0],
                                vals[SpsToUsaskSolarSceneInterface.EXPECTED_SOLAR_RADIUS].values[scene_index],
                                vals[SpsToUsaskSolarSceneInterface.WAVELENGTHS].values[scene_index],
                                vals[SpsToUsaskSolarSceneInterface.DETECTOR_BORESIGHT].values,
                                vals[SpsToUsaskSolarSceneInterface.START_BORESIGHT_POINTING].values[scene_index],
                                vals[SpsToUsaskSolarSceneInterface.END_BORESIGHT_POINTING].values[scene_index],
                                vals[SpsToUsaskSolarSceneInterface.INTEGRATION_TIMES].values[scene_index])


class SpsToUsaskLunarSceneInterface:
    """
    Class representing the output of the SPS measurements of the moon for the PRNU lunar
    calibration.  It contains a set of images of the moon at a handful of locations on the
    detector for a single wavelength, in netcdf file format.

    In truth, this class may only represent the output from the SPS given to the calibration
    processes, in which case

    Main fields:

    - Images - units of DN
        - dim 0: image index
        - dim 1: image vertical dimension
        - dim 2: image horizontal dimension
    - detector resolution and size
    - center wavelength
    - detector pixel solid angles
    - detector boresight
    - other data tbd
    - lunar disk parameters:
        - lunar disk center, radius
        - etc.
    """

    # Dimensions:
    NUM_VERT_PIXELS = 'num_vert_pixels'
    NUM_HORZ_PIXELS = 'num_horz_pixels'
    PIXEL_LOCATION_SIZE = 'pixel_location_size'
    CAL_WL_DIM = 'calibration_wavelength_dimension'
    XYZ = 'xyz'
    DETECTOR_POS_DIM = 'detector_positions_dim'

    # Variable names:
    DETECTOR_BORESIGHT = 'detector_boresight'
    SCENES = 'scenes'
    TIMES = 'times'
    EXPECTED_LUNAR_CENTER = 'expected_lunar_center'
    EXPECTED_LUNAR_RADIUS = 'expected_lunar_radius'
    SAT_TO_MOON_DISTANCE = 'satellite_to_moon_distance'
    DETECTOR_FOV_DEG = 'detector_fov_deg'
    CALIBRATION_WAVELENGTHS = 'calibration_wavelengths_nm'
    START_SNAPSHOT_BORESIGHT_POINTING = 'start_snapshot_pointing'
    END_SNAPSHOT_BORESIGHT_POINTING = 'end_snapshot_pointing'
    INTEGRATION_TIMES = 'integration_times'
    SUBSOLAR_SELEN_LONGITUDE = 'subsolar_selenographic_longitude'
    SUBSAT_SELEN_LONGITUDE = 'subsatellite_selenographic_longitude'
    SUBSAT_SELEN_LATITUDE = 'subsatellite_selenographic_latitude'
    LUNAR_PHASE_ANGLE_DEG = 'lunar_phase_angle'
    SUN_TO_MOON_DISTANCE_AU = 'sun_to_moon_distance_au'
    SATELLITE_POSITION_ECEF = 'satellite_position_ecef'
    DETECTOR_POSITIONS = 'detector_lunar_pixel_positions'

    def __init__(self):
        self._read_nc_file = None
        self._write_nc_file = None

    def open_file_for_write(self,
                            filename: str,
                            calibration_wavelengths: np.ndarray,
                            detector_fov: Tuple[float, float],
                            detector_size_pixels: Tuple[int, int],
                            detector_boresight_pixels: Tuple[float, float],
                            grid_pixel_positions: np.ndarray):
        self._write_nc_file = nc.Dataset(filename, mode='w', format='NETCDF4')

        # Dimensions:
        self._num_vert_pixels = self._write_nc_file.createDimension(self.NUM_VERT_PIXELS, detector_size_pixels[0])
        self._num_horz_pixels = self._write_nc_file.createDimension(self.NUM_HORZ_PIXELS, detector_size_pixels[1])
        self._pixel_location_size = self._write_nc_file.createDimension(self.PIXEL_LOCATION_SIZE, 2)
        self._calibration_wl_dim = self._write_nc_file.createDimension(self.CAL_WL_DIM, len(calibration_wavelengths))
        self._boresight_vector_dim = self._write_nc_file.createDimension(self.XYZ, 3)
        self._exposure_detector_positions_dim = self._write_nc_file.createDimension(self.DETECTOR_POS_DIM,
                                                                                    len(grid_pixel_positions))

        # Variables:
        self._satellite_position = self._write_nc_file.createVariable(self.SATELLITE_POSITION_ECEF, float,
                                                                      (self.DETECTOR_POS_DIM, self.CAL_WL_DIM,
                                                                       self.XYZ))

        self._exposure_detector_positions = self._write_nc_file.createVariable(self.DETECTOR_POSITIONS, int,
                                                                               (self.DETECTOR_POS_DIM,
                                                                                self.PIXEL_LOCATION_SIZE))

        self._exposure_detector_positions[:] = grid_pixel_positions
        self._exposure_detector_positions.description = 'Approximate center pixel locations of the snapshots ' \
                                                        '- limited to a small set across the detector'
        self._exposure_detector_positions.units = 'Vertical pixel number, horizontal pixel number'

        self._snapshot_pointing_start = self._write_nc_file.createVariable(self.START_SNAPSHOT_BORESIGHT_POINTING,
                                                                           float,
                                                                           (self.DETECTOR_POS_DIM, self.CAL_WL_DIM,
                                                                            self.XYZ))
        self._snapshot_pointing_start.units = 'Arbitrary'
        self._snapshot_pointing_start.description = 'ECEF absolute vector describing the pointing of the instrument ' \
                                                    'boresight at the beginning of each integration period'

        self._snapshot_pointing_end = self._write_nc_file.createVariable(self.END_SNAPSHOT_BORESIGHT_POINTING,
                                                                         float,
                                                                         (self.DETECTOR_POS_DIM, self.CAL_WL_DIM,
                                                                          self.XYZ))
        self._snapshot_pointing_end.units = 'Arbitrary'
        self._snapshot_pointing_end.description = 'ECEF absolute vector describing the pointing of the instrument ' \
                                                  'boresight at the end of each integration period'

        self._integration_times = self._write_nc_file.createVariable(self.INTEGRATION_TIMES, float,
                                                                     (self.DETECTOR_POS_DIM, self.CAL_WL_DIM,))
        self._integration_times.units = 'seconds'
        self._integration_times.description = 'Exposure time of each snapshot of the sun'

        self._detector_boresight = self._write_nc_file.createVariable(self.DETECTOR_BORESIGHT, float,
                                                                      (self.PIXEL_LOCATION_SIZE,))
        self._detector_boresight.units = 'Pixels'
        self._detector_boresight[:] = [detector_boresight_pixels[0], detector_boresight_pixels[1]]

        self._calibration_wavelengths = self._write_nc_file.createVariable(self.CALIBRATION_WAVELENGTHS, float,
                                                                           (self.CAL_WL_DIM,))
        self._calibration_wavelengths[:] = calibration_wavelengths
        self._calibration_wavelengths.units = 'nm'
        self._calibration_wavelengths.description = 'Wavelength in nm for which the set of images is intended ' \
                                                    'to calibrate'

        self._detector_fov = self._write_nc_file.createVariable(self.DETECTOR_FOV_DEG, float,
                                                                (self.PIXEL_LOCATION_SIZE,))
        self._detector_fov.units = 'deg'
        self._detector_fov.description = 'Vertical and horizontal angular width of detector, respectively'
        self._detector_fov[:] = [detector_fov[0], detector_fov[1]]

        self._scenes = self._write_nc_file.createVariable(self.SCENES, float,
                                                          (self.DETECTOR_POS_DIM, self.CAL_WL_DIM,
                                                           self.NUM_VERT_PIXELS, self.NUM_HORZ_PIXELS))
        self._scenes.units = 'DN'
        self._scenes.description = 'Scene at detector resolution'

        self._times = self._write_nc_file.createVariable(self.TIMES, float,
                                                         (self.DETECTOR_POS_DIM, self.CAL_WL_DIM,))
        self._times.units = 'MJD'

        self._expected_lunar_center = self._write_nc_file.createVariable(self.EXPECTED_LUNAR_CENTER, float,
                                                                         (self.DETECTOR_POS_DIM, self.CAL_WL_DIM,
                                                                          self.PIXEL_LOCATION_SIZE))
        self._expected_lunar_center.description = 'Center of the lunar disk in vertical and horizontal pixels, ' \
                                                  'respectively'
        self._expected_lunar_center.units = 'Pixels'

        self._expected_lunar_radius = self._write_nc_file.createVariable(self.EXPECTED_LUNAR_RADIUS, float,
                                                                         (self.DETECTOR_POS_DIM, self.CAL_WL_DIM,))
        self._expected_lunar_radius.units = 'Pixels'

        self._satellite_to_moon_distance = self._write_nc_file.createVariable(self.SAT_TO_MOON_DISTANCE, float,
                                                                              (self.DETECTOR_POS_DIM,
                                                                               self.CAL_WL_DIM,))
        self._satellite_to_moon_distance.units = 'km'

        # SUBSOLAR_SELEN_LONGITUDE = 'subsolar_selenographic_longitude'
        # SUBSAT_SELEN_LONGITUDE = 'subsatellite_selenographic_longitude'
        # SUBSAT_SELEN_LATITUDE = 'subsatellite_selenographic_latitude'
        # LUNAR_PHASE_ANGLE_DEG = 'lunar_phase_angle'
        # SUN_TO_MOON_DISTANCE_AU = 'sun_to_moon_distance_au'

        self._subsolar_selen_longitude = self._write_nc_file.createVariable(self.SUBSOLAR_SELEN_LONGITUDE, float,
                                                                            (self.DETECTOR_POS_DIM,
                                                                             self.CAL_WL_DIM,))
        self._subsolar_selen_longitude.units = 'Degrees'
        self._subsolar_selen_longitude.description = 'Selenographic longitude of the sun projected on to the moon'

        self._subsat_selen_latitude = self._write_nc_file.createVariable(self.SUBSAT_SELEN_LATITUDE, float,
                                                                         (self.DETECTOR_POS_DIM, self.CAL_WL_DIM,))
        self._subsat_selen_latitude.units = 'Degrees'
        self._subsat_selen_latitude.description = 'Selenographic latitude of the satellite projected onto the moon'

        self._subsat_selen_longitude = self._write_nc_file.createVariable(self.SUBSAT_SELEN_LONGITUDE, float,
                                                                          (self.DETECTOR_POS_DIM,
                                                                           self.CAL_WL_DIM,))
        self._subsat_selen_longitude.units = 'Degrees'
        self._subsat_selen_longitude.description = 'Selenographic longitude of the satellite projected onto the moon'

        self._lunar_phase_angle = self._write_nc_file.createVariable(self.LUNAR_PHASE_ANGLE_DEG, float,
                                                                     (self.DETECTOR_POS_DIM, self.CAL_WL_DIM,))
        self._lunar_phase_angle.units = 'Degrees'
        self._lunar_phase_angle.description = 'Phase angle in degrees of the moon'

        self._sun_to_moon_distance_au = self._write_nc_file.createVariable(self.SUN_TO_MOON_DISTANCE_AU, float,
                                                                           (self.DETECTOR_POS_DIM,
                                                                            self.CAL_WL_DIM,))
        self._sun_to_moon_distance_au.units = 'Astronomical Units'

    def get_channel(self, config: altius_config.AltiusL1ppInflightConfig):
        vals = self._read_nc_file.variables
        wl = vals[SpsToUsaskLunarSceneInterface.CALIBRATION_WAVELENGTHS][0].values
        return config.get_channel(wl)

    def write_next_lunar_scene_to_netcdf(self,
                                         time: np.datetime64,
                                         scene: np.ndarray,
                                         wavelength: float,
                                         satellite_position: np.ndarray,
                                         expected_lunar_center: Tuple[float, float],
                                         expected_lunar_radius: float,
                                         sat_moon_distance_km: float,
                                         start_boresight_pointing: np.ndarray,
                                         end_boresight_pointing: np.ndarray,
                                         integration_time: float,
                                         subsolar_selen_longitude,
                                         subsat_selen_latitude,
                                         subsat_selen_longitude,
                                         sun_to_moon_distance_au,
                                         lunar_phase_angle_deg):
        # Determine the closest detector positions and wavelengths:
        wl_idx = np.argmin(np.abs(wavelength - self._calibration_wavelengths))
        pos_idx = np.argmin(
            np.sqrt(np.sum((np.asarray(expected_lunar_center) - self._exposure_detector_positions) ** 2, axis=1)))

        self._satellite_position[pos_idx, wl_idx, :] = satellite_position
        self._times[pos_idx, wl_idx] = mjd_conv.datetime64_to_mjd(time)
        self._scenes[pos_idx, wl_idx, :, :] = scene
        self._expected_lunar_center[pos_idx, wl_idx, :] = [expected_lunar_center[0], expected_lunar_center[1]]
        self._expected_lunar_radius[pos_idx, wl_idx] = expected_lunar_radius
        self._satellite_to_moon_distance[pos_idx, wl_idx] = sat_moon_distance_km
        self._snapshot_pointing_start[pos_idx, wl_idx, :] = start_boresight_pointing
        self._snapshot_pointing_end[pos_idx, wl_idx, :] = end_boresight_pointing
        self._integration_times[pos_idx, wl_idx] = integration_time
        self._subsolar_selen_longitude[pos_idx, wl_idx] = subsolar_selen_longitude
        self._subsat_selen_latitude[pos_idx, wl_idx] = subsat_selen_latitude
        self._subsat_selen_longitude[pos_idx, wl_idx] = subsat_selen_longitude
        self._sun_to_moon_distance_au[pos_idx, wl_idx] = sun_to_moon_distance_au
        self._lunar_phase_angle[pos_idx, wl_idx] = lunar_phase_angle_deg

    def close_write_nc(self):
        self._write_nc_file.close()

    def close_read_nc(self):
        if self._read_nc_file is not None:
            self._read_nc_file.close()

    def open_read_dataset(self, filename):
        self._read_nc_file = xr.open_dataset(filename)  # , chunks={SpsToUsaskSceneInterface.TIME_DIM: 4})

    def get_num_scenes(self):
        return self._read_nc_file.dims[SpsToUsaskSolarSceneInterface.TIME_DIM]

    def get_wavelengths(self):
        return self._read_nc_file[SpsToUsaskLunarSceneInterface.CALIBRATION_WAVELENGTHS].values

    def load_scene(self, wavelength, detector_position_index) -> AltiusLunarScene:
        vals = self._read_nc_file.variables
        k = detector_position_index

        wl_idx = np.argmin(wavelength - vals[SpsToUsaskLunarSceneInterface.CALIBRATION_WAVELENGTHS])

        expected_center = vals[SpsToUsaskSolarSceneInterface.EXPECTED_SOLAR_CENTER][k, wl_idx].values
        return AltiusLunarScene(vals[SpsToUsaskLunarSceneInterface.SCENES][k, wl_idx].values,
                                vals[SpsToUsaskLunarSceneInterface.DETECTOR_FOV_DEG].values,
                                vals[SpsToUsaskLunarSceneInterface.SAT_TO_MOON_DISTANCE][k, wl_idx].values,
                                expected_center[1],
                                expected_center[0],
                                vals[SpsToUsaskLunarSceneInterface.EXPECTED_LUNAR_RADIUS][k, wl_idx].values,
                                vals[SpsToUsaskLunarSceneInterface.CALIBRATION_WAVELENGTHS][wl_idx].values,
                                vals[SpsToUsaskLunarSceneInterface.DETECTOR_BORESIGHT].values,
                                vals[SpsToUsaskLunarSceneInterface.START_SNAPSHOT_BORESIGHT_POINTING][k, wl_idx].values,
                                vals[SpsToUsaskLunarSceneInterface.END_SNAPSHOT_BORESIGHT_POINTING][k, wl_idx].values,
                                vals[SpsToUsaskLunarSceneInterface.INTEGRATION_TIMES][k, wl_idx].values,
                                vals[SpsToUsaskLunarSceneInterface.SUN_TO_MOON_DISTANCE_AU][k, wl_idx].values,
                                vals[SpsToUsaskLunarSceneInterface.LUNAR_PHASE_ANGLE_DEG][k, wl_idx].values,
                                vals[SpsToUsaskLunarSceneInterface.SUBSAT_SELEN_LATITUDE][k, wl_idx].values,
                                vals[SpsToUsaskLunarSceneInterface.SUBSAT_SELEN_LONGITUDE][k, wl_idx].values,
                                vals[SpsToUsaskLunarSceneInterface.SUBSOLAR_SELEN_LONGITUDE][k, wl_idx].values)

    def get_num_pixels_scene(self):
        return self._num_vert_pixels, self._num_horz_pixels


class LunarPrnuSpotCheckOutput:
    """
        This class describes the interface to be provided to the QinetiQ SPS as a result of the
        output of the spot check PRNU calibration using the moon as a source.  This output contains
        the calibration result for all chosen calibration wavelengths in a channel (uv, vis, nir).
        The wavelengths are set as the discrete set of wavelengths where the ROLO model parameters
        are well known.

        For now, this consists of the following netcdf structure.

        Dimensions:

        - NUM_VERT_PIXELS
            - value: 'num_vert_pixels'
            - fixed number of vertical pixels on the detector
        - NUM_HORZ_PIXELS
            - value: 'num_horz_pixels'
            - fixed number of horizontal pixels on the detector
        - WAVELENGTH_DIM
            - value: 'wavelength_dimension'
            - Number of wavelengths corresponding to the exact ROLO database wavelengths for this channel
        - SCENE_DIM
            - value: 'scene_dimension'
            - Index corresponding to one of the handful of moon center positions on the detector for each spot check
        - PIXEL_LOCATION_SIZE
            - value: 'pixel_location_size'
            - (vertical, horizontal) pixel position on the detector

        Variables:

        - CALIBRATION_WAVELENGTHS
            - value: 'calibration_wavelengths'
            - size: [WAVELENGTH_DIM,]
            - wavelength, in nm, of each PRNU calibration
        - RATIO
            - value: 'lunar_scene_to_rolo_irradiance_ratio'
            - size: [WAVELENGTH_DIM, SCENE_DIM]
            - The resulting ratio of the measured irradiance compared to the expected irradiance from the ROLO model
        - RATIO_ERROR
            - value: 'lunar_scene_to_rolo_irradiance_ratio_error'
            - size: [WAVELENGTH_DIM, SCENE_DIM]
            - The error in the resulting ratio of the measured irradiance compared to the expected irradiance
              from the ROLO model

        - PIXEL_MASK
            - value: 'pixel_mask'
            - size: [WAVELENGTH_DIM, SCENE_DIM, NUM_VERT_PIXELS, NUM_HORZ_PIXELS]
            - Mask of the detector for each measurement where the moon illuminated the pixels.

        - EXPOSURE_CENTER_PIXEL_POSITIONS
            - value: 'exposure_center_pixel_positions'
            - size: [SCENE_DIM, PIXEL_LOCATION_SIZE]
            - Center pixel positions of where the moon is expected to be for each measurement
        """

    # Dimensions:
    NUM_VERT_PIXELS = 'num_vert_pixels'
    NUM_HORZ_PIXELS = 'num_horz_pixels'
    WAVELENGTH_DIM = 'wavelength_dimension'
    SCENE_DIM = 'scene_dimension'
    PIXEL_LOCATION_SIZE = 'pixel_location_size'

    # Variables:
    RATIO = 'lunar_scene_to_rolo_irradiance_ratio'
    RATIO_ERROR = 'lunar_scene_to_rolo_irradiance_ratio_error'
    PIXEL_MASK = 'pixel_mask'
    CALIBRATION_WAVELENGTHS = 'calibration_wavelengths'
    EXPOSURE_CENTER_PIXEL_POSITIONS = 'exposure_center_pixel_positions'

    def __init__(self, filename, num_vert_pixels, num_horz_pixels,
                 wavelengths: np.ndarray, grid_pixel_positions: np.ndarray):
        """
        Parameters
        ----------
        filename: str

        num_vert_pixels: int

        num_horz_pixels: int

        wavelengths: np.ndarray
            Calibration wavelengths used for this channel

        grid_pixel_positions: np.ndarray
            Array [len(wavelengths), 2] of (vertical, horizontal) pixel positions where the moon
            is expected to be imaged on the detector
        """
        self._write_nc_file = nc.Dataset(filename, mode='w', format='NETCDF4')

        # Dimensions:
        self._wavelength_dim = self._write_nc_file.createDimension(self.WAVELENGTH_DIM,
                                                                   len(wavelengths))  # variable size (appending)
        self._scene_dim = self._write_nc_file.createDimension(self.SCENE_DIM, len(grid_pixel_positions))
        self._num_vert_pixels = self._write_nc_file.createDimension(self.NUM_VERT_PIXELS, num_vert_pixels)
        self._num_horz_pixels = self._write_nc_file.createDimension(self.NUM_HORZ_PIXELS, num_horz_pixels)
        self._pixel_location_size = self._write_nc_file.createDimension(self.PIXEL_LOCATION_SIZE, 2)

        # Variables:
        self._exposure_detector_positions = self._write_nc_file.createVariable(self.EXPOSURE_CENTER_PIXEL_POSITIONS,
                                                                               int,
                                                                               (self.SCENE_DIM,
                                                                                self.PIXEL_LOCATION_SIZE))
        self._exposure_detector_positions[:] = grid_pixel_positions
        self._wavelengths = self._write_nc_file.createVariable(self.CALIBRATION_WAVELENGTHS, float,
                                                               (self.WAVELENGTH_DIM,))
        self._wavelengths[:] = wavelengths
        self._irradiance_ratio = self._write_nc_file.createVariable(self.RATIO, float,
                                                                    (self.WAVELENGTH_DIM, self.SCENE_DIM))
        self._irradiance_ratio_error = self._write_nc_file.createVariable(self.RATIO_ERROR, float,
                                                                          (self.WAVELENGTH_DIM, self.SCENE_DIM))
        self._pixel_mask = self._write_nc_file.createVariable(self.PIXEL_MASK, int,
                                                              (self.WAVELENGTH_DIM, self.SCENE_DIM,
                                                               self.NUM_VERT_PIXELS, self.NUM_HORZ_PIXELS))

    def write_next_ratio_and_mask(self, irradiance_ratio: float, irradiance_ratio_error: float,
                                  pixel_mask: np.ndarray, expected_lunar_center: Tuple[float, float],
                                  wavelength: float):
        # Determine the closest detector positions and wavelengths:
        wl_idx = np.argmin(np.abs(wavelength - self._wavelengths))
        pos_idx = np.argmin(
            np.sqrt(np.sum((np.asarray(expected_lunar_center) - self._exposure_detector_positions) ** 2, axis=1)))

        self._irradiance_ratio[wl_idx, pos_idx] = irradiance_ratio
        self._irradiance_ratio_error[wl_idx, pos_idx] = irradiance_ratio_error

        self._pixel_mask[wl_idx, pos_idx, :, :] = pixel_mask

    def close(self):
        self._write_nc_file.close()


class SolarPrnuMapOutput:
    """
    This class describes the interface to be provided to the QinetiQ SPS as a result of the
    output of the PRNU calibration.

    For now, this consists of the following netcdf structure.

    Dimensions:

    - NUM_VERT_PIXELS
        - value: 'num_vert_pixels'
        - fixed number of vertical pixels on the detector
    - NUM_HORZ_PIXELS
        - value: 'num_horz_pixels'
        - fixed number of horizontal pixels on the detector
    - CAL_WL
        - value: 'cal_wl'
        - the length of the number of PRNU images, for each calibration wavelength

    Variables:

    - CALIBRATION_WAVELENGTHS
        - value: 'calibration_wavelengths'
        - size: [CAL_WL,]
        - wavelength, in nm, of each PRNU calibration
    - PRNU_MAP
        - value: 'prnu_map'
        - size: [CAL_WL, NUM_VERT_PIXELS, NUM_HORZ_PIXELS]
        - the output PRNU map for each calibration wavelength, on a unitary scale
    - PRNU_MAP_ERROR
        - value: 'prnu_map_error'
        - size: [CAL_WL, NUM_VERT_PIXELS, NUM_HORZ_PIXELS]
        - the error in each pixel in each PRNU map
    """
    # Dimensions:
    NUM_VERT_PIXELS = 'num_vert_pixels'
    NUM_HORZ_PIXELS = 'num_horz_pixels'
    CAL_WL = 'cal_wl'

    # Variables:
    CALIBRATION_WAVELENGTHS = 'calibration_wavelengths'
    PRNU_MAP = 'prnu_map'
    PRNU_MAP_ERROR = 'prnu_map_error'

    def __init__(self,
                 filename: str,
                 calibration_wavelength: float,
                 prnu_image: np.ndarray,
                 prnu_image_error: np.ndarray):
        """
        Parameters
        ----------
        filename: str
            Absolute file path of the output netcdf.

        calibration_wavelength: float
            Wavelength of the PRNU calibration in nm.

        prnu_image: np.ndarray
            The obtained PRNU image for the detector.  The PRNU image is obtained by correcting the
            solar images for limb darkening, stitching them together, and dividing by the ideal modelled
            and limb-darkening-corrected solar radiance.  The ideal image is a 2D array of 1.0's.

        prnu_image_error: np.ndarray
            The error in the PRNU for each detector pixel.
        """

        self._write_nc_file = nc.Dataset(filename, mode='w', format='NETCDF4')

        nv, nh = np.shape(prnu_image)

        # Dimensions:
        self._num_vert_pixels = self._write_nc_file.createDimension(self.NUM_VERT_PIXELS, nv)
        self._num_horz_pixels = self._write_nc_file.createDimension(self.NUM_HORZ_PIXELS, nh)
        self._cal_wl_dim = self._write_nc_file.createDimension(self.CAL_WL, 1)

        # Variables:
        self._calibration_wavelength = self._write_nc_file.createVariable(self.CALIBRATION_WAVELENGTHS, float,
                                                                          (self.CAL_WL,))
        self._calibration_wavelength.units = 'nm'
        self._calibration_wavelength[:] = [calibration_wavelength]
        self._calibration_wavelength.description = 'Wavelength in nm for which the set of images is intended ' \
                                                   'to calibrate'

        self.prnu_map = self._write_nc_file.createVariable(self.PRNU_MAP, float,
                                                           (self.NUM_VERT_PIXELS, self.NUM_HORZ_PIXELS))
        self.prnu_map.units = 'photons/s/cm2/sr/nm'
        self.prnu_map.description = 'PRNU Flat-field map after applying corrections and removing old PRNU'
        self.prnu_map[:, :] = prnu_image

        self.prnu_map_error = self._write_nc_file.createVariable(self.PRNU_MAP_ERROR, float,
                                                                 (self.NUM_VERT_PIXELS, self.NUM_HORZ_PIXELS))
        self.prnu_map_error.units = 'photons/s/cm2/sr/nm'
        self.prnu_map_error.description = 'Error in PRNU Flat-field map'
        self.prnu_map_error[:, :] = prnu_image_error

        self._write_nc_file.close()
