from __future__ import absolute_import, division, print_function

import glob
import logging
import os
from datetime import datetime
from enum import Enum

from sktime import mjd as mjd_conv
import drms
import numpy as np
from astropy.io import fits
from sasktran.solarspectrum import SolarSpectrum

from altius_l1pp_inflight.configuration import altius_config
from altius_l1pp_inflight.scene.disk_geometry import D_1AU, SolarGeometry, SOLAR_RADIUS, center_solar_disk
from altius_l1pp_inflight.scene.generation.solar.calibration_util import SolarCalibrationUtil
from altius_l1pp_inflight.scene.limbdark_mu_pierceandslaughter1977 import LimbDarkening_PS1977
from altius_l1pp_inflight.srf.SrfInterface import SrfInterface

HORIZONTAL_HMI_PIXEL_INDEX, VERTICAL_HMI_PIXEL_INDEX = np.meshgrid(np.arange(0, 4096) + 0.5, np.arange(0, 4096) + 0.5)

sao = SolarSpectrum(name='SAO2010')


class SolarDiskRadianceResult:
    """
    Simple container class for the results of high resolution solar disk radiance arrays.
    The primary ability the class offers is to scale the resulting disk radiance values
    to the correct values based off the irradiance of the SAO2010 spectrum, and corrected
    for limb darkening at the desired wavelength.
    """

    def __init__(self, radiance_disk: np.ndarray,
                 vertical_disk_pixel_centroid: float,
                 horizontal_disk_pixel_centroid: float,
                 average_radius_pixels: float,
                 actual_solar_distance_km: float,
                 mu: np.ndarray,
                 time: datetime,
                 wavelength: float = 617.3,
                 solar_radius_apparent_angle_deg: float = 0.25):
        self.radiance_disk: np.ndarray = radiance_disk  # arbitrary units
        self._v_c: float = vertical_disk_pixel_centroid
        self._h_c: float = horizontal_disk_pixel_centroid
        self._average_radius_pixels: float = average_radius_pixels
        self._actual_solar_distance: float = actual_solar_distance_km
        self._mu: np.ndarray = mu
        self._time: datetime = time
        self._wavelength = wavelength
        self._solar_radius_apparent_angle_deg = solar_radius_apparent_angle_deg

    def radiance(self):
        return self.radiance_disk

    def center_pixel(self):
        return self._h_c, self._v_c

    def average_solar_radius_in_pixels(self):
        return self._average_radius_pixels

    def angular_radius_degrees(self):
        return self._solar_radius_apparent_angle_deg

    def actual_solar_distance(self):
        return self._actual_solar_distance

    def mu_solar_disk(self):
        return self._mu

    def time(self) -> datetime:
        return self._time

    def to_scaled_image_result(self, wavelength: float, srf: SrfInterface,
                               horizontal_pixel_shift=0, vertical_pixel_shift=0):
        """
        For a requested wavelength, SRF, and solar disk detector position, scales the solar disk
        radiance appropriately.  The integral of the pixel radiance across the solar disk
        needs to add up to the SAO2010 solar spectrum irradiance computed at the central wavelength and with
        a specified SRF.

        This function corrects for:
            * viewing angle: this is encapsulated in the horizontal_pixel_shift and vertical_pixel_shift
            parameters - this helps to determine the correct differential solid angles of pixel to perform the
            correct integration across the disk
            * limb darkening: employs the Pierce & Slaughter limb darkening model to remove the limb darkening
            at the current wavelength and apply the limb darkening at the requested wavelength
            * SRF: computes the total irradiance that would be observed from the solar disk using the SAO2010
            solar spectrum and with any spectral response function at the requested wavelength

        Parameters
        ----------
        wavelength : float
           Wavelength in nanometers

        srf : SrfInterface
            Spectral response function at wavelengths in srf_wavelen.  The srf is essentially the weighting
            factor for the SAO2010 solar spectrum.

        horizontal_pixel_shift: float
            Shift in virtual pixels at the HMI image resolution to the left of the boresight on the detector

        vertical_pixel_shift: float
            Shift in virtual pixels at the HMI image resolution above the boresight on the detector


        Returns
        -------
        SolarDiskRadianceResult
           An instance of the class with updated parameters for the wavelength, look angle, and +srf used
        """
        logging.info('Scaling radiance to correct for SRF, SAO2010 irradiance, viewing angle, and limb darkening.')
        logging.info('wavelength: ', wavelength)
        logging.info('horizontal_pixel_shift: ', horizontal_pixel_shift)
        logging.info('vertical_pixel_shift: ', vertical_pixel_shift)

        logging.info('Applying limb darkening correction')
        radiance_disk_wl = self._correct_disk_for_limb_darkening(wavelength)

        logging.info('Applying solar irradiance and solid angle correction')
        # photons/s/cm2/nm/sr:
        srf_corrected = self._correct_for_solar_irradiance(horizontal_pixel_shift, vertical_pixel_shift,
                                                           srf, radiance_disk_wl, wavelength)

        return SolarDiskRadianceResult(srf_corrected, self._v_c, self._h_c,
                                       self._average_radius_pixels, self._actual_solar_distance,
                                       self._mu, self._time,
                                       wavelength, self._solar_radius_apparent_angle_deg)

    def _correct_disk_for_limb_darkening(self, wavelength):
        limb_darkening = LimbDarkening_PS1977()
        mu_profile = np.linspace(0, 1.0, 50)
        ld_original_profile = limb_darkening.limb_darkening(self._wavelength, mu_profile)
        ld_old_wl = np.interp(self._mu, mu_profile, ld_original_profile)
        flattened_radiance = self.radiance_disk / ld_old_wl

        ld_wl_profile = limb_darkening.limb_darkening(wavelength, mu_profile)
        ld_wl = np.interp(self._mu, mu_profile, ld_wl_profile)
        radiance_disk_wl = flattened_radiance * ld_wl
        return radiance_disk_wl

    def _correct_for_solar_irradiance(self, horizontal_pixel_shift,
                                      vertical_pixel_shift,
                                      srf: SrfInterface,
                                      radiance_disk_wl, wavelength):
        """
            Computes the integrated disk irradiance to match the solar irradiance,
            and returns the photons/s/cm2/sr/nm in each pixel, accounting for each
            pixel's solid angle
        """
        # integrated irradiance of image in [arbitrary units]
        horizontal_width = len(self.radiance_disk[0])
        vertical_width = len(self.radiance_disk)
        fov_v = vertical_width * self._solar_radius_apparent_angle_deg / self._average_radius_pixels
        fov_h = horizontal_width * self._solar_radius_apparent_angle_deg / self._average_radius_pixels
        solar_geometry = SolarGeometry(horizontal_width, vertical_width,
                                       -vertical_pixel_shift, -horizontal_pixel_shift,
                                       fov_v, fov_h)
        solar_geometry.set_disk(self._average_radius_pixels, self._h_c, self._v_c, self._actual_solar_distance)
        integrated_disk_irradiance, _, delta_omega = solar_geometry.compute_irradiance_for_viewing_angle(
            self.radiance_disk)

        logging.info('Computed current disk irradiance: ', integrated_disk_irradiance)

        # SAO irradiance: [photons/s/cm2/nm]
        solar_cal = SolarCalibrationUtil(solar_geometry, srf)
        sao_irrad = solar_cal.compute_solar_irradiance(wavelength, self._actual_solar_distance)

        logging.info('Actual solar irradiance using SRF at ', wavelength, ' nm: ', sao_irrad, ' photons/s/nm/cm2')
        logging.info('Scaling disk pixel radiances by factor: ', sao_irrad / integrated_disk_irradiance)

        absolute_radiance_disk_wl = radiance_disk_wl * sao_irrad / integrated_disk_irradiance / delta_omega
        return absolute_radiance_disk_wl  # [photons/s/cm2/sr/nm]

    def wavelength(self):
        return self._wavelength


def least_squares_2d(Z: np.ndarray, notes: str):
    ny, nx = np.shape(Z)
    ZZ = Z / np.max(Z)
    x = np.arange(0, nx)
    y = np.arange(0, ny)
    X, Y = np.meshgrid(x, y, copy=False)
    Xf = X.flatten()
    Yf = Y.flatten()

    # A = np.array([Xf*0+1, Xf, Yf, Xf**2, Xf**2*Yf, Xf**2*Yf**2, Yf**2, Xf*Yf**2, Xf*Yf]).T
    A = np.array([Xf * 0 + 1, Xf, Yf, Xf * Yf]).T
    B = ZZ.flatten()

    coeff, r, rank, s = np.linalg.lstsq(A, B)

    fitfunc = coeff[0] + X * coeff[1] + Y * coeff[2] + X * Y * coeff[
        3]  # X**2*coeff[3] + X**2*Y*coeff[4] + X**2*Y**2*coeff[5] + X*Y**2*coeff[6] + X*Y*coeff[7]

    print('Least squares fit for central flattened region in ', notes, ':')
    print('   1: ', coeff[0])
    print('   X: ', coeff[1])
    print('   Y: ', coeff[2])
    print('  X*Y: ', coeff[3])

    return coeff, r, rank, s, fitfunc


class HmiSunspotScenario(Enum):
    NO_SUNSPOTS = 0
    ONE_SUNSPOT = 1
    TWO_SUNSPOTS = 2
    SEVERAL_SUNSPOTS = 3

    def to_string_directory(self):

        if self is HmiSunspotScenario.NO_SUNSPOTS:
            sunspot_dir = 'no_sunspots'
        elif self is HmiSunspotScenario.ONE_SUNSPOT:
            sunspot_dir = 'one_sunspot'
        elif self is HmiSunspotScenario.TWO_SUNSPOTS:
            sunspot_dir = 'two_sunspots'
        elif self is HmiSunspotScenario.SEVERAL_SUNSPOTS:
            sunspot_dir = 'several_sunspots'
        else:
            raise AttributeError('Scenario not supported')

        return sunspot_dir

    def as_select_statement(self):

        if self is HmiSunspotScenario.NO_SUNSPOTS:
            tsel = '2019.11.14_12:30:00_TAI/60m'
        elif self is HmiSunspotScenario.ONE_SUNSPOT:
            tsel = '2012.10.23_00:00:00_TAI/60m'
        elif self is HmiSunspotScenario.TWO_SUNSPOTS:
            tsel = '2016.04.05_00:00:00_TAI/60m'
        elif self is HmiSunspotScenario.SEVERAL_SUNSPOTS:
            tsel = '2014.02.14_00:00:00_TAI/60m'
        else:
            raise AttributeError('Scenario not supported')

        return tsel


def compute_solar_angular_radius_deg(actual_solar_distance):
    solar_radius_apparent_angle_deg = np.rad2deg(np.arctan(SOLAR_RADIUS / actual_solar_distance))
    return solar_radius_apparent_angle_deg


class HmiFileManager:
    """
    This class submits a data export request using the 'fits'
    protocol and downloads the requested files. Note that the 'as-is'
    protocol should be used instead of 'fits', if record keywords in the FITS
    headers are not needed, as it greatly reduces the server load.

    See https://github.com/sunpy/drms/tree/master/examples for more

    This example requires a registered export email address. You can register
    JSOC exports at: http://jsoc.stanford.edu/ajax/register_email.html
    """

    def __init__(self):

        self._email = 'klj588@mail.usask.ca'  # registered on website, no need to enter a different one

        # Use 'as-is' instead of 'fits', if record keywords are not needed in the
        # FITS header. This greatly reduces the server load!
        self._export_protocol = 'fits'

        # Series, harpnum, timespan and segment selection
        self._series = 'hmi.Ic_720s'  # continuum

        self._segments = ['continuum']

        # Download directory
        config = altius_config.AltiusL1ppInflightConfig()
        self._home_dir = config.get_hmi_home_directory()  # os.environ[SOLAR_HMI_DIR_ENVIRON]

    def input_file_home_directory(self):
        return self._home_dir

    def _download_scenario_files(self, scenario_key: HmiSunspotScenario, tsel_custom_override=None,
                                 outdir_custom_override=None):
        """
        Downloads the pre-determined files for a specified sunspot scenario.  Downloads 5-6 files
        which were measured at roughly 12 minute intervals.  The files are meant to cover the time
        period of roughly half an orbit.

        Parameters
        ----------
        scenario_key : HmiSunspotScenario
           Enum corresponding to one of the four scenarios: no sunspots, one sunspot, two sunspots,
           and several sunspots, each corresponding to a different measurement time in the HMI database.

        tsel_custom_override : str
            Custom select statement for choosing files at a user-specified time (mostly for tests).  Overrides the
            directory pulled from scenario_key

        outdir_custom_override : str
            Overrides the output directory for the files


        Returns
        -------
        Nothing.  If files are not present on the host machine specified by the ALTIUS_L1PP_INFLIGHT_DATABASE_FOLDER
        environment variable, then the method queries the HMI database and downloads the files.
        """
        out_dir = os.path.join(self._home_dir, 'continuum', scenario_key.to_string_directory())
        if outdir_custom_override is not None:
            out_dir = outdir_custom_override

        # Create download directory if it does not exist yet.
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)

        # Create DRMS client, use debug=True to see the query URLs.
        client = drms.Client(verbose=True)

        # Check if the email address was set at the top of this script. If not, ask for
        # a registered email address.
        if not self._email or not client.check_email(self._email):
            raise RuntimeError('Email address is not valid or not registered.')

        tsel = scenario_key.as_select_statement()
        if tsel_custom_override is not None:
            tsel = tsel_custom_override

        # Data export query string
        # qstr = '%s[%d][%s]{%s}' % (series, harpnum, tsel, ','.join(segments))
        qstr = '%s[%s]{%s}' % (self._series, tsel, ','.join(self._segments))
        logging.info('Data export query:\n  %s\n' % qstr)

        # Submit export request using the 'fits' protocol
        logging.info('Submitting export request...')
        r = client.export(qstr, method='url', protocol=self._export_protocol, email=self._email)

        # Print request URL.
        logging.info('\nRequest URL: %s' % r.request_url)
        logging.info('%d file(s) available for download:\n' % len(r.urls))

        # Download selected files.
        r.download(out_dir)
        logging.info('Download finished.')
        logging.info('\nDownload directory:\n  "%s"\n' % os.path.abspath(out_dir))

    def _load_fits(self, filename: str) -> SolarDiskRadianceResult:
        """
        Load FITS files using the necessary query protocol

        Parameters
        ----------
        filename: string
            Full path to the FITS file

        Returns
        -------
        result: Dictionary with fields:
            'radiance_disk_617_hmi'
            'x_c'
            'y_c'
            'average_radius_pixels'
            'actual_solar_distance'
            'mu'
            'datetime'
        """
        logging.info('Loading FITS file: ', filename)
        # ext = os.path.splitext(filename)  # Check the file extension
        path_dir, fname = os.path.split(filename)
        time_str = str.split(fname[12:], sep='_')
        year = int(time_str[0][0:4])
        month = int(time_str[0][4:6])
        day = int(time_str[0][6:8])
        hour = int(time_str[1][0:2])
        minute = int(time_str[1][2:4])
        second = int(time_str[1][4:6])

        time = datetime(year, month, day, hour, minute, second)
        logging.info('Extracted time: ', time)

        # ok = ok and (ext[1] == '.fits')  # and if its a fits file
        with fits.open(filename) as hdul:  # and open the FITS file
            hdul.verify('silentfix+warn')  # Fix up the FITS non-standard issues quietly
            data = hdul[1].data  # and get the data

        data[np.isnan(data)] = 0

        logging.info('Determining centroid by applying sobel operator to find edges, and finding the max')
        logging.info('The assumption is that the space-to-disk transition gives much larger gradients than sunspots')
        average_radius_pixels, x_c, y_c, alg_pts = center_solar_disk(data)
        # don't really need the additional arguments, as we know what the HMI data looks like

        # Since there are some edges detection inside/outside the disk, clean up by finding the most likely radius
        disk_x, disk_y = np.where((HORIZONTAL_HMI_PIXEL_INDEX - x_c) ** 2 +
                                  (VERTICAL_HMI_PIXEL_INDEX - y_c) ** 2 <= average_radius_pixels ** 2)
        # all the disk pixels, including darks
        data_mask = np.ones_like(HORIZONTAL_HMI_PIXEL_INDEX, dtype=bool)
        data_mask[disk_x.flatten(), disk_y.flatten()] = False
        radiance_disk_617_hmi = np.ma.masked_array(data=data, mask=data_mask)

        time_mjd = mjd_conv.datetime_to_mjd(time)
        actual_irrad = sao.irradiance(617.3, mjd=time_mjd)
        irrad_1au = sao.irradiance(617.3)

        actual_solar_distance = np.sqrt(irrad_1au[0] / actual_irrad[0]) * D_1AU
        solar_radius_apparent_angle_deg = compute_solar_angular_radius_deg(actual_solar_distance)
        logging.info('solar radius apparent angle: ', solar_radius_apparent_angle_deg)

        solar_geometry = SolarGeometry(4096, 4096, y_c, x_c, 0.5, 0.5)  # fov doesn't matter
        solar_geometry.set_disk(average_radius_pixels, x_c, y_c, actual_solar_distance)
        mu = solar_geometry.mu()

        return SolarDiskRadianceResult(radiance_disk=radiance_disk_617_hmi,
                                       vertical_disk_pixel_centroid=y_c, horizontal_disk_pixel_centroid=x_c,
                                       average_radius_pixels=average_radius_pixels,
                                       actual_solar_distance_km=actual_solar_distance, mu=mu,
                                       time=time, wavelength=617.3,
                                       solar_radius_apparent_angle_deg=solar_radius_apparent_angle_deg)

    def load_scenario_files_as_images(self, scenario: HmiSunspotScenario) -> np.ndarray:
        """
        Returns loaded in images + auxillary data ordered by the time of measurement
        for a given sunspot scenario

        This interface is defined in the _load_fits function

        Parameters
        ----------
        scenario : HmiSunspotScenario
            Enum describing the sunspot scenario (none, one, two, or several)

        Returns
        -------
        all_results : np.ndarray of dictionary
            Each array entry is a dictionary defined in the _load_fits function

        """
        print('Loading all files for scenario: ', scenario.to_string_directory())
        sunspot_dir = scenario.to_string_directory()
        scenario_directory = os.path.join(self._home_dir, 'continuum', sunspot_dir)
        ok = os.path.exists(scenario_directory)

        if not ok:
            print('Data not available for scenario ', scenario.to_string_directory(), ' downloading files to ',
                  scenario_directory)
            self._download_scenario_files(scenario)

        files = glob.glob(os.path.join(scenario_directory, '*.fits'))

        all_results = []
        mjds = []

        for f in files:
            result = self._load_fits(f)
            all_results.append(result)
            date = result.time()
            mjds.append(mjd_conv.datetime_to_mjd(date))

        mjds = np.array(mjds)
        all_results = np.array(all_results)

        all_results = all_results[np.argsort(mjds)]

        return all_results
