import os
from typing import Sequence, Union, List

import numpy as np

# ---------------------------------------------------------------------------------
# Table IV from Pierce and Slaughter 1976. SOLAR LIMB DARKENING
# ndl303: 2020-04-18. I have manually checked each value to verify they are correct
# Coefficients of 5th degree, mu = cos(theta) fit to the limb darkening, Equation (11)
# Lambda    No  A(5)      B(5)      C(5)      D(5)      E(5)      F(5)   Pe x 104
# --------------------------------------------------------------------------------
os.chdir(os.path.dirname(os.path.abspath(__file__)))
g_pierce_and_slaughter_mu_table = np.genfromtxt('g_mu_table.csv', delimiter=',')


# ------------------------------------------------------------------------------
#        Class LimbDarkening_PS1977
# ------------------------------------------------------------------------------
class LimbDarkening_PS1977:
    """
    Calculates limb-darkening for the Sun using polynomial coefficients for :math:`\mu` from table IV  of
    Pierce and Slaughter 1977. SOLAR LIMB DARKENING I: :math:`\lambda` (3033-7297), Solar Physics 51 (1977) 25-41
    """

    def __init__(self):
        global g_pierce_and_slaughter_mu_table
        rawtable = np.array(g_pierce_and_slaughter_mu_table)

        # Fetch the wavelength from the Pierce and Slaughter 1977 Table IV table and convert to nanometers
        self._wavelength: np.ndarray = rawtable[:, 0] / 10.0
        self._wavindex: np.ndarray = np.argsort(
            self._wavelength)  # Generate the list of sorted indices. This is required for np.searchsorted.
        self._number: np.ndarray = rawtable[:, 1]  # get the number of points
        # Extract the limb darkening coefficients and flip so the higherst polynomial is first.
        # This is required for numpy.polyval
        self._coeffs: np.ndarray = np.fliplr(rawtable[:, 2:8])
        self._pe: np.ndarray = rawtable[:, 8] / 1.0E4

    # ------------------------------------------------------------------------------
    #        wavelengths
    # ------------------------------------------------------------------------------

    @property
    def wavelengths(self) -> np.ndarray:
        """
        Returns the array of wavelengths in nanometers used in the Pierce and Slaughter 1977
        measurements.
        """
        return self._wavelength

    def limb_darkening_integral(self, wavelength_nm: float, start_mu: float = 0.0, end_mu: float = 1.0) -> float:
        """
        Computes the integral of the limb darkening polynomial for the range [start_mu, end_mu].
        Default is for the entire solar disk, but generally can be computed for a ring within
        the disk of the solar disk.

        Parameters
        ----------
        wavelength_nm : float
           The wavelength in nm at which limb darkening will be calculated. This must be a scalar.

        start_mu : float, Optional
           Starting value for the integral - 0.0 corresponds to the edge of the disk

        end_mu : float, Optional
           Ending value for the integral - 1.0 corresponds to the middle of the disk.

        Returns
        -------
        integral : float
           The resulting area under the limb darkening curve from start_mu to end_mu
        """
        wl_idx0, wl_idx1 = self._find_bounding_wavelength_indices(wavelength_nm)

        integral_start = np.zeros((2,))
        integral_end = np.zeros((2,))

        c1 = self._coeffs[wl_idx1, :]
        c0 = self._coeffs[wl_idx0, :]

        c11 = [a / (len(c1) - i) for i, a in enumerate(c1)]
        c00 = [a / (len(c0) - i) for i, a in enumerate(c0)]

        integral_start[1] = np.polyval(c11, start_mu)
        integral_start[0] = np.polyval(c00, start_mu)

        integral_end[1] = np.polyval(c11, end_mu)
        integral_end[0] = np.polyval(c00, end_mu)

        w = np.array([self._wavelength[wl_idx0], self._wavelength[wl_idx1]])
        int_start = np.interp(wavelength_nm, w, integral_start)
        int_end = np.interp(wavelength_nm, w, integral_end)

        return int_end - int_start

    # ------------------------------------------------------------------------------
    #        limb_darkening
    # ------------------------------------------------------------------------------

    def limb_darkening(self, wavelength_nm: float, usermu: Union[np.ndarray, List[float], float]) -> np.ndarray:
        """
        Calculates the limb darkening at a given wavelength for an array of :math:`\mu` ( :math:`\cos\theta` ) values.
        The code will linear interpolate curves between the wavelengths measured by Pierce and Slaughter but it
        will not extrapolate.

        Parameters
        ----------
        wavelength_nm : float
           The wavelength in nm at which limb darkening will be calculated. This must be a scalar.

        usermu : float, Sequence or array
           The set of :math:`\mu` values at which the the limb-darkening is required. This can be a scalar, sequence or
           array. Note that if the fractional distance from the center of the sun is :math:`x` then :math:`\mu`
           is given by :math:`\sqrt{1 - x^2}` where :math:`-1 \geq x \leq 1`.

        Returns
        -------
        np.ndarray
           A 1-D array the same size as `usermu`.
        """

        idx0, idx1 = self._find_bounding_wavelength_indices(wavelength_nm)

        if type(usermu) is np.ndarray:
            mu = usermu
        elif isinstance(usermu, Sequence):
            mu = np.array(usermu)
        else:
            mu = np.array([usermu])

        darkvals = np.zeros([mu.size, 2])

        darkvals[:, 1] = np.polyval(self._coeffs[idx1, :], mu)
        darkvals[:, 0] = np.polyval(self._coeffs[idx0, :], mu)

        w = np.array([self._wavelength[idx0], self._wavelength[idx1]])
        darkening = np.zeros_like(mu)

        # TODO just extrapolating here for the limb darkening below 303 nm - NEED a better placeholder at some point
        for idx in range(mu.size):
            darkening[idx] = np.interp(wavelength_nm, w, darkvals[idx, :], right=np.nan)
        return darkening

    def _find_bounding_wavelength_indices(self, wavelength_nm):
        idx1 = np.searchsorted(self._wavelength, wavelength_nm, sorter=self._wavindex)
        if idx1 == 0:
            idx1 = 1
        if idx1 == self._wavelength.size:
            idx1 = self._wavelength.size - 1
        idx0 = idx1 - 1
        return idx0, idx1
