import numpy as np
from scipy.ndimage import black_tophat


def detect_sunspots(image, threshold_value=0.3):
    """
        Values near the edge of the sun tend to give false positives, so it is up to the user
        to pass in a reasonable portion of the sun that avoids the edges.

        Additionally, passing in a flattened sun removes the possibility of detecting darkened edges
        as sunspots.

        Source: Watson and Fletcher (2011), "Automated sunspot detection and the evolution of
        sunspot magnetic fields during solar cycle 23"
    """
    inverted_image = 1 / image
    inverted_image /= np.nanmax(inverted_image)
    inverted_image *= np.nanmax(image)

    sunspot_removed = np.copy(image)

    sunspot_removed[np.isnan(sunspot_removed)] = 0.0  # replace background values with 0's

    # look for different sizes
    for fw in [0.001, 0.005, 0.01, 0.02, 0.03, 0.04]:
        filter_width = int(len(image) * fw)
        filtered_black2020 = black_tophat(inverted_image, (filter_width, filter_width))
        background_removed = inverted_image - filtered_black2020
        threshold_upper = (1.0 + threshold_value) * np.nanmean(background_removed)
        threshold_lower = (1.0 - threshold_value) * np.nanmean(background_removed)
        background_removed[np.isnan(background_removed)] = np.nanmean(background_removed)
        sunspot_removed[background_removed > threshold_upper] = np.nan
        sunspot_removed[background_removed < threshold_lower] = np.nan
        # filtered_black2020[np.isnan(filtered_black2020)] = 0.0
        # sunspot_removed[filtered_black2020 > threshold_upper] = np.nan
        # sunspot_removed[filtered_black2020 < threshold_lower] = np.nan
        # inverted_image[filtered_black2020 > threshold_upper] = np.nan
        # inverted_image[filtered_black2020 < threshold_lower] = np.nan

    return inverted_image, filtered_black2020, background_removed, sunspot_removed
