import pickle

import matplotlib.font_manager as font_manager
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr
from matplotlib.colors import BoundaryNorm
from matplotlib.ticker import MaxNLocator
from sasktran.solarspectrum import SolarSpectrum

# Set the font dictionaries (for plot title and axis titles)
from altius_l1pp_inflight.srf import uv_srf, wavelen_registration

title_font = {'name': 'Arial', 'size': '20', 'color': 'black', 'weight': 'normal',
              'verticalalignment': 'bottom'}  # Bottom vertical alignment for more space
axis_font = {'name': 'Arial', 'size': '18'}

# Set the font properties (for use in legend)
font_path = 'C:\\Windows\\Fonts\\Arial.ttf'
font_prop = font_manager.FontProperties(fname=font_path, size=16)

directory = 'A:\\USASK\\ALTIUS_DATA\\l1pp_datafiles\\+srf'

srf = uv_srf.AltiusUVSrf(directory)

solar_spectrum = SolarSpectrum(name='SAO2010')
res = 0.01
wavelength_grid = np.linspace(202, 400, int(198 / res))

sao = solar_spectrum.irradiance(wavelength_grid)
solar = xr.DataArray(sao, dims={'wavelength'}, coords={'wavelength': wavelength_grid})

assume_gaussian = True

snr_vals = np.logspace(1, 4, 41)

num_wavelen = 1001
wavelengths = None

spectrum_convergence = np.zeros((len(snr_vals), num_wavelen))
fwhm_convergence = np.zeros((len(snr_vals), num_wavelen))
analysis_wavelengths = np.linspace(250, 350, num_wavelen)
for k, snr in enumerate(snr_vals):
    print(f'Computing fit for SNR={snr}')
    fit_fwhm, actual_fwhm, WL, sim_spectrum, fit_spectrum, rel_percent_diff = \
        wavelen_registration.fit_using_gaussian_approximation(srf, analysis_wavelengths,
                                                              assume_gaussian, snr=snr)

    wavelengths = WL

    # Percent difference of FWHM
    a_f = actual_fwhm.sel(wavelength=WL)
    actual_fwhm_wavenumber = a_f.values * 1e-7 * (1 / (a_f.wavelength.values * 1e-7)) ** 2
    wavenum_error = 0.05 * actual_fwhm_wavenumber
    fit_fwhm_wavenumber = fit_fwhm * 1e-7 * (1 / (WL * 1e-7)) ** 2
    fwhm_perc_diff = 100 * (fit_fwhm_wavenumber - actual_fwhm_wavenumber) / actual_fwhm_wavenumber
    fwhm_convergence[k] = fwhm_perc_diff

    # Percent difference between fitted and actual observation
    spectrum_convergence[k] = rel_percent_diff

# for pcolormesh, need edges of grid, not center points (add one point)
wl_res = wavelengths[1] - wavelengths[0]
wavelength_edges = wavelengths - 0.5 * wl_res
np.append(wavelength_edges, wavelength_edges[-1] + wl_res)

# for snr, need to work in log space
log_snr = np.log(snr_vals)
log_snr_res = log_snr[1] - log_snr[0]
log_snr_edges = log_snr - 0.5 * log_snr_res
np.append(log_snr_edges, log_snr_edges[-1] + log_snr_res)
snr_edges = np.exp(log_snr_edges)

# save to file
with open('fwhm_fit_with_noise_no_shift.pickle', 'wb') as f:
    pickle.dump({'wavelength_edges': wavelength_edges,
                 'snr_edges': snr_edges,
                 'spectrum_convergence': spectrum_convergence,
                 'fwhm_convergence': fwhm_convergence}, f)

levels = MaxNLocator(nbins=15).tick_values(np.min(spectrum_convergence), np.max(spectrum_convergence))
cmap = plt.get_cmap('PiYG')
norm = BoundaryNorm(levels, ncolors=cmap.N, clip=True)

fig1, ax1 = plt.subplots()
im = ax1.pcolormesh(wavelength_edges, snr_edges, spectrum_convergence)
fig1.colorbar(im, ax=ax1)
ax1.set_title('Percent Convergence of Spectrum with Varying Noise', **title_font)
ax1.set_xlabel('Wavelength [nm]', **axis_font)
ax1.set_ylabel('SNR', **axis_font)
ax1.set_yscale('log')
xticklabels = ax1.get_xticklabels()
yticklabels = ax1.get_yticklabels()
ax1.set_xticklabels(xticklabels, **axis_font)
ax1.set_yticklabels(yticklabels, **axis_font)

fig2, ax2 = plt.subplots()
im = ax2.pcolormesh(wavelength_edges, snr_edges, fwhm_convergence)
fig2.colorbar(im, ax=ax2)
ax2.set_title('Percent Convergence of Fit FWHM with Varying Noise', **title_font)
ax2.set_xlabel('Wavelength [nm]', **axis_font)
ax2.set_ylabel('SNR', **axis_font)
ax2.set_yscale('log')
ax2.set_xticklabels(xticklabels, **axis_font)
ax2.set_yticklabels(yticklabels, **axis_font)

plt.show()
