import unittest
from datetime import datetime

import matplotlib.patches as patches
import matplotlib.pyplot as plt
import numpy as np

from altius_l1pp_inflight.scene.disk_geometry import center_solar_disk, D_1AU
from altius_l1pp_inflight.scene.generation.solar.hmi_files import SolarDiskRadianceResult
from altius_l1pp_inflight.scene.generation.solar.virtual_solar_detector import RasterScanGrid, VirtualDetectorPixelArray


def test_partially_off_detector(show_figures=False):
    solar_pixel_radius = 900
    rg = RasterScanGrid(solar_radius_angular_size_deg=0.25,
                        scan_rate_deg_per_sec=1.0,
                        time_per_image_sec=1.0,
                        fov_horizontal_deg=2.0,
                        fov_vertical_deg=2.0,
                        v_in=8,
                        h_in=8)

    vp = VirtualDetectorPixelArray(disk_radius_pixels=solar_pixel_radius,
                                   solar_scan_grid=rg,
                                   boresight_horz_portion=0.0,
                                   boresight_vert_portion=0.0)
    disk_radiance = np.zeros((2000, 2000))
    x, y = np.meshgrid(np.arange(0, 2000), np.arange(0, 2000))
    r = np.sqrt((x - 1000) ** 2 + (y - 1000) ** 2)
    disk_radiance[r <= solar_pixel_radius] = 1.0

    old_sum = np.nansum(disk_radiance)
    solar_image = SolarDiskRadianceResult(disk_radiance, 1000, 1000, solar_pixel_radius, 149.6e6, np.empty((1, 1)),
                                          datetime.now())
    vp.set_current_raster_image(solar_image, 1000, 1000, 8)
    low_res, image_geometry = vp.to_altius_resolution_image()

    bbox = image_geometry['raster_bounding_box']

    params = {'distance_to_sun_km': D_1AU, 'detector_angular_width_degrees': (2.0, 2.0)}
    rr, px, py, ap = center_solar_disk(low_res, **params)

    if show_figures:
        fig, ax = plt.subplots()
        ax.imshow(disk_radiance)

        fig, ax = plt.subplots()
        ax.imshow(low_res)
        ax.add_patch(patches.Rectangle((bbox[0], bbox[1]), bbox[2], bbox[3], fill=False, edgecolor='blue'))
        plt.show()
    # plt.plot([650, 950], [400, 400], 'r-')

    return px, py, bbox[0] + 0.5 * bbox[2], bbox[1] + 0.5 * bbox[3]


def test_conserve_photons(show_figures=False):
    solar_pixel_radius = 900
    rg = RasterScanGrid(solar_radius_angular_size_deg=0.25,
                        scan_rate_deg_per_sec=1.0,
                        time_per_image_sec=1.0,
                        fov_horizontal_deg=2.0,
                        fov_vertical_deg=2.0,
                        v_in=8,
                        h_in=8)

    vp = VirtualDetectorPixelArray(disk_radius_pixels=solar_pixel_radius,
                                   solar_scan_grid=rg,
                                   boresight_horz_portion=0.0,
                                   boresight_vert_portion=0.0)
    disk_radiance = np.zeros((2000, 2000))
    x, y = np.meshgrid(np.arange(0, 2000), np.arange(0, 2000))
    r = np.sqrt((x - 1000) ** 2 + (y - 1000) ** 2)
    r1 = np.sqrt((x - 1200) ** 2 + (y - 800) ** 2)
    disk_radiance[r <= solar_pixel_radius] = 1.0
    disk_radiance[r1 <= 300] = 2.0

    old_sum = np.nansum(disk_radiance)
    solar_image = SolarDiskRadianceResult(disk_radiance, 1000, 1000, solar_pixel_radius,
                                          149.6e6, np.empty((1, 1)),
                                          datetime.now())
    vp.set_current_raster_image(solar_image, 1000, 1000, 25)
    low_res, image_geometry = vp.to_altius_resolution_image()

    bbox = image_geometry['raster_bounding_box']

    new_sum = np.nansum(low_res)
    old_sum *= vp._vert_scale * vp._horz_scale

    print('old sum: ', old_sum)
    print('new sum: ', new_sum)

    if show_figures:
        fig, ax = plt.subplots()
        ax.imshow(disk_radiance)

        fig, ax = plt.subplots()
        ax.imshow(low_res)
        ax.add_patch(patches.Rectangle((bbox[0], bbox[1]), bbox[2], bbox[3], fill=False, edgecolor='blue'))
        plt.plot([650, 950], [400, 400], 'r-')

        h0 = bbox[0]
        v0 = bbox[1]
        hw = int(bbox[2])
        vw = int(bbox[3])
        plt.figure()
        plt.title('Low resolution - bounding box')
        plt.imshow(low_res[v0:v0 + vw, h0:h0 + hw])
        plt.figure()
        plt.plot(low_res[400, 650:950])
        plt.show()

    return old_sum, new_sum


class TestDownsampling(unittest.TestCase):

    def test_conserve_photons_after_downsampling_image(self):
        old_sum, new_sum = test_conserve_photons()
        self.assertAlmostEqual(old_sum, new_sum,
                               msg='Difference in total photons is greater than 0.1%',
                               delta=0.001 * old_sum)

    def test_expected_center_when_solar_disk_partially_off_detector(self):
        px, py, ax, ay = test_partially_off_detector()
        self.assertAlmostEqual(px, ax,
                               msg='Expected xcenter: {}, actual: {}'.format(ax, px),
                               delta=1.0)
        self.assertAlmostEqual(py, ay,
                               msg='Expected ycenter: {}, actual: {}'.format(ay, py),
                               delta=1.0)


if __name__ == '__main__':
    unittest.main()
