import unittest
from datetime import datetime

import matplotlib.pyplot as plt
import numpy as np

from altius_l1pp_inflight.scene.generation.solar.hmi_files \
    import VERTICAL_HMI_PIXEL_INDEX, HORIZONTAL_HMI_PIXEL_INDEX, SolarDiskRadianceResult


def test_solid_angle_computation(vertical_radii_offset, horizontal_radii_offset):
    n = len(VERTICAL_HMI_PIXEL_INDEX[:, 0])
    radiance = np.ones((n, n))

    solar_radius = 696340
    x_c = (n - 1) / 2
    y_c = (n - 1) / 2
    average_radius_pixels = n / 8

    # Compute the differential solid angles at an off-axis angle
    vertical_offset = vertical_radii_offset * average_radius_pixels
    horizontal_offset = horizontal_radii_offset * average_radius_pixels

    image_pixel_radii = np.sqrt((HORIZONTAL_HMI_PIXEL_INDEX - x_c) ** 2 + (VERTICAL_HMI_PIXEL_INDEX - y_c) ** 2)
    pixel_to_radius_scale = solar_radius / average_radius_pixels
    image_pixel_radii *= pixel_to_radius_scale
    fractional_radii = image_pixel_radii / solar_radius
    mu = np.sqrt(1 - fractional_radii ** 2)

    solar_distance = 147.43e6
    result = SolarDiskRadianceResult(radiance_disk=radiance,
                                     vertical_disk_pixel_centroid=y_c, horizontal_disk_pixel_centroid=x_c,
                                     average_radius_pixels=average_radius_pixels,
                                     actual_solar_distance_km=solar_distance,
                                     mu=mu,
                                     time=datetime(2019, 11, 14, 12, 35, 0, 0),
                                     wavelength=410.1,
                                     solar_radius_apparent_angle_deg=0.265)

    # Pick one where we're staring straight at the sun
    delta_omega, phi, theta = result.compute_solid_angles(horizontal_pixel_shift=0,
                                                          vertical_pixel_shift=0)

    plt.figure()
    plt.imshow(delta_omega)
    plt.title('Differential Solid Angle Over Detector')
    plt.colorbar()
    plt.show()

    sum_solid_angle_expected = np.pi * solar_radius ** 2 / solar_distance ** 2

    # sum over the disk pixels
    sum_solid_angle_actual = np.sum(np.where(fractional_radii <= 1.0,
                                             delta_omega,
                                             np.zeros_like(image_pixel_radii)))

    # make sure it's within 5%
    is_close = np.isclose(sum_solid_angle_expected, sum_solid_angle_actual,
                          rtol=0.001)

    if int(is_close) == 0:
        raise AssertionError('Actual sum ' + str(sum_solid_angle_actual) +
                             ' not within 5% of expected value: ' +
                             str(sum_solid_angle_expected))
    else:
        print('Boresight disk differential angles test passed')


def test_scaling_limb_darkening():
    pass


def test_scaling_srf():
    pass


def test_scaling_viewing_angle():
    pass


def test_scaling_total_disk_irradiance():
    pass


class HmiSolarDiskTests(unittest.TestCase):

    def test_solid_angle_computation_straight_view(self):
        test_solid_angle_computation(0, 0)

    def test_solid_angle_computation_off_axis_view(self):
        test_solid_angle_computation(-4, 10)

    def test_scaling_limb_darkening(self):
        test_scaling_limb_darkening()

    def test_scaling_srf(self):
        test_scaling_srf()

    def test_scaling_viewing_angle(self):
        test_scaling_viewing_angle()

    def test_scaling_total_disk_irradiance(self):
        test_scaling_total_disk_irradiance()


if __name__ == '__main__':
    test_solid_angle_computation(0, 0)
    test_solid_angle_computation(-4, 10)
    test_scaling_limb_darkening()
    test_scaling_srf()
    test_scaling_viewing_angle()
    test_scaling_total_disk_irradiance()
