import os
import unittest

import numpy as np
import sktime.mjd as mjd_conv
import xarray as xr

from altius_l1pp_inflight.scene.altius_scene import SpsToUsaskSolarSceneInterface, SolarPrnuMapOutput
from altius_l1pp_inflight.scene.disk_geometry import D_1AU


class TestSolarInterfaces(unittest.TestCase):

    def test_solar_sps_scene_interface(self):
        scene = SpsToUsaskSolarSceneInterface()
        nv = 60
        nh = 60
        wavelength = 500
        positions = np.array([[10, 20], [10, 40],
                              [30, 20], [30, 40],
                              [50, 20], [50, 40]])

        scene.open_file_for_write('temp_scene.nc',
                                  wavelength,
                                  (2.0, 2.0),
                                  (nv, nh),
                                  (nv / 2, nh / 2))

        r = 7.5
        dv, dh = np.meshgrid(np.arange(0, nv), np.arange(0, nh))

        expected_scenes = np.zeros((len(positions), nv, nh))
        expected_sat_posn = np.zeros((len(positions), 3))
        expected_start_bs = np.copy(expected_sat_posn)
        expected_end_bs = np.copy(expected_sat_posn)
        expected_integration_times = np.zeros((len(positions)))
        expected_times = np.zeros_like(expected_integration_times)
        sat_to_sun = D_1AU * np.ones_like(expected_integration_times)
        expected_centers = np.zeros((len(positions), 2))
        wavelengths = []

        for pos_idx, pos in enumerate(positions):
            wavelengths.append(wavelength)
            start_bs = np.random.random((3,))
            expected_start_bs[pos_idx] = start_bs
            end_bs = np.random.random((3,))
            expected_end_bs[pos_idx] = end_bs
            t = np.random.rand() * 1.5
            expected_integration_times[pos_idx] = t
            sat_to_sun[pos_idx] = D_1AU

            posn = pos + np.random.rand()
            expected_centers[pos_idx] = posn

            R = np.sqrt((dv - posn[0]) ** 2 + (dh - posn[1]) ** 2)
            next_scene = R <= r
            next_scene = next_scene.astype(int) * np.random.randint(255)
            expected_scenes[pos_idx] = next_scene

            expected_times[pos_idx] = mjd_conv.datetime64_to_mjd(np.datetime64('2019-11-14'))

            scene.write_next_solar_scene_to_netcdf(np.datetime64('2019-11-14'),
                                                   next_scene,
                                                   wavelength,
                                                   posn,
                                                   r,
                                                   D_1AU,
                                                   start_bs, end_bs,
                                                   t)
        scene.close_write_nc()

        ds = xr.open_dataset('temp_scene.nc')
        vals = ds.data_vars.variables
        self.assertListEqual(vals[SpsToUsaskSolarSceneInterface.WAVELENGTHS].values.tolist(),
                             wavelengths)
        self.assertListEqual(vals[SpsToUsaskSolarSceneInterface.SAT_TO_SUN_DISTANCE].values.tolist(),
                             sat_to_sun.tolist())
        self.assertListEqual(vals[SpsToUsaskSolarSceneInterface.TIMES].values.tolist(),
                             expected_times.tolist())
        self.assertListEqual(vals[SpsToUsaskSolarSceneInterface.SCENES].values.tolist(),
                             expected_scenes.tolist())
        self.assertListEqual(vals[SpsToUsaskSolarSceneInterface.EXPECTED_SOLAR_CENTER].values.tolist(),
                             expected_centers.tolist())
        self.assertListEqual(vals[SpsToUsaskSolarSceneInterface.START_BORESIGHT_POINTING].values.tolist(),
                             expected_start_bs.tolist())
        self.assertListEqual(vals[SpsToUsaskSolarSceneInterface.END_BORESIGHT_POINTING].values.tolist(),
                             expected_end_bs.tolist())
        # TODO something weird is happening with integration times:
        # self.assertListEqual(vals[SpsToUsaskSolarSceneInterface.INTEGRATION_TIMES].values.tolist(),
        #                      expected_integration_times.tolist())

        ds.close()
        os.remove('temp_scene.nc')

    def test_solar_prnu_output_interface(self):
        nv = 60
        nh = 60
        wavelength = 450

        expected_prnu = np.random.random((nv, nh))
        expected_prnu_error = np.random.random((nv, nh)) * 0.01

        SolarPrnuMapOutput('temp.nc', wavelength, expected_prnu, expected_prnu_error)

        ds = xr.open_dataset('temp.nc')
        vals = ds.data_vars.variables
        self.assertListEqual(vals[SolarPrnuMapOutput.PRNU_MAP].values.tolist(),
                             expected_prnu.tolist())
        self.assertListEqual(vals[SolarPrnuMapOutput.PRNU_MAP_ERROR].values.tolist(),
                             expected_prnu_error.tolist())

        ds.close()
        os.remove('temp.nc')
